from typing import Any, Type

import pytest
from cryptography.hazmat.primitives.asymmetric import ed448, ed25519, x448, x25519

from jwskate import Jwk, JwsCompact, OKPJwk, UnsupportedOKPCurve


@pytest.mark.parametrize("curve", ["Ed25519", "Ed448", "X25519", "X448"])
def test_jwk_okp_generate(curve: str) -> None:
    jwk = OKPJwk.generate(crv=curve, kid="myokpkey")
    assert jwk.kty == "OKP"
    assert jwk.kid == "myokpkey"
    assert "x" in jwk
    assert "d" in jwk

    assert jwk.supported_encryption_algorithms() == []


def test_okp_ed25519_sign() -> None:
    jwk = Jwk(
        {
            "kty": "OKP",
            "crv": "Ed25519",
            "d": "nWGxne_9WmC6hEr0kuwsxERJxWl7MmkZcDusAxyuf2A",
            "x": "11qYAYKxCrfVS_7TyWQHOg7hcvPapiMlrwIaaPcHURo",
        }
    )
    assert isinstance(jwk, OKPJwk)
    assert jwk.is_private
    payload = "Example of Ed25519 signing".encode()

    jws = JwsCompact.sign(payload, jwk=jwk, alg="EdDSA")
    assert jws.alg == "EdDSA"
    assert jws.headers == {"alg": "EdDSA"}
    assert jws.payload == payload
    assert (
        jws
        == "eyJhbGciOiJFZERTQSJ9.RXhhbXBsZSBvZiBFZDI1NTE5IHNpZ25pbmc.hgyY0il_MGCjP0JzlnLWG1PPOt7-09PGcvMg3AIbQR6dWbhijcNR4ki4iylGjg5BhVsPt9g7sVvpAr_MuM0KAg"
    )

    assert jws.verify_signature(jwk=jwk.public_jwk(), alg="EdDSA")


def test_unknown_curve() -> None:
    with pytest.raises(UnsupportedOKPCurve):
        Jwk({"kty": "OKP", "crv": "foobar", "x": "abcd"})


@pytest.mark.parametrize(
    "crv,private_key_class,public_key_class",
    [
        ("Ed25519", ed25519.Ed25519PrivateKey, ed25519.Ed25519PublicKey),
        ("Ed448", ed448.Ed448PrivateKey, ed448.Ed448PublicKey),
        ("X448", x448.X448PrivateKey, x448.X448PublicKey),
        ("X25519", x25519.X25519PrivateKey, x25519.X25519PublicKey),
    ],
)
def test_from_to_cryptography(
    crv: str, private_key_class: Type[Any], public_key_class: Type[Any]
) -> None:
    private_key = private_key_class.generate()
    private_jwk = Jwk(private_key)
    assert private_jwk.kty == "OKP"
    assert private_jwk.crv == crv
    assert private_jwk.is_private
    cryptography_private_key = private_jwk.cryptography_key
    assert isinstance(cryptography_private_key, private_key_class)

    public_jwk = Jwk(private_key.public_key())
    assert public_jwk.kty == "OKP"
    assert public_jwk.crv == crv
    assert not public_jwk.is_private
    cryptography_public_key = public_jwk.cryptography_key
    assert isinstance(cryptography_public_key, public_key_class)



"""This module implements the JWE Compact format."""

import warnings
from typing import Any, Dict, Mapping, Optional, Type, Union

from backports.cached_property import cached_property
from binapy import BinaPy

from jwskate.jwa import (
    BasePbes2,
    Pbes2_HS256_A128KW,
    Pbes2_HS384_A192KW,
    Pbes2_HS512_A256KW,
)
from jwskate.jwk import Jwk, SymmetricJwk
from jwskate.jwk.alg import UnsupportedAlg
from jwskate.token import BaseCompactToken


class InvalidJwe(ValueError):
    """Raised when an invalid JWE token is parsed."""


class JweCompact(BaseCompactToken):
    """Represents a Json Web Encryption object, as defined in RFC7516."""

    def __init__(self, value: Union[bytes, str]):
        """Initialize a Jwe based on its compact representation.

        Args:
            value: the compact representation for this Jwe
        """
        super().__init__(value)

        if self.value.count(b".") != 4:
            raise InvalidJwe(
                "A JWE must contain a header, an encrypted key, an IV, a ciphertext and an authentication tag, separated by dots"
            )

        header, cek, iv, ciphertext, auth_tag = self.value.split(b".")
        try:
            self.headers = BinaPy(header).decode_from("b64u").parse_from("json")
            self.additional_authenticated_data = header
        except ValueError:
            raise InvalidJwe(
                "Invalid JWE header: it must be a Base64URL-encoded JSON object"
            )

        try:
            self.wrapped_cek = BinaPy(cek).decode_from("b64u")
        except ValueError:
            raise InvalidJwe(
                "Invalid JWE cek: it must be a Base64URL-encoded binary data (bytes)"
            )

        try:
            self.initialization_vector = BinaPy(iv).decode_from("b64u")
        except ValueError:
            raise InvalidJwe(
                "Invalid JWE iv: it must be a Base64URL-encoded binary data (bytes)"
            )

        try:
            self.ciphertext = BinaPy(ciphertext).decode_from("b64u")
        except ValueError:
            raise InvalidJwe(
                "Invalid JWE ciphertext: it must be a Base64URL-encoded binary data (bytes)"
            )

        try:
            self.authentication_tag = BinaPy(auth_tag).decode_from("b64u")
        except ValueError:
            raise InvalidJwe(
                "Invalid JWE authentication tag: it must be a Base64URL-encoded binary data (bytes)"
            )

    @classmethod
    def from_parts(
        cls,
        *,
        headers: Mapping[str, Any],
        cek: bytes,
        iv: bytes,
        ciphertext: bytes,
        tag: bytes,
    ) -> "JweCompact":
        """Initialize a JWE from its different parts (header, cek, iv, ciphertext, tag).

        Args:
          headers: the headers (as a mapping of name: value)
          cek: the raw cek
          iv: the raw iv
          ciphertext: the raw ciphertext
          tag: the authentication tag

        Returns:
            the initialized JweCompact instance
        """
        return cls(
            b".".join(
                (
                    BinaPy.serialize_to("json", headers).to("b64u"),
                    BinaPy(cek).to("b64u"),
                    BinaPy(iv).to("b64u"),
                    BinaPy(ciphertext).to("b64u"),
                    BinaPy(tag).to("b64u"),
                )
            )
        )

    @cached_property
    def alg(self) -> str:
        """Return the alg from the JWE header.

        The `alg` header contains the Key Management algorithm used to wrap or derive the CEK.

        Returns:
            the alg value

        Raises:
            AttributeError: if there is no alg header or it is not a str
        """
        alg = self.get_header("alg")
        if alg is None or not isinstance(alg, str):
            raise AttributeError("This JWE doesn't have a valid 'alg' header")
        return alg

    @cached_property
    def enc(self) -> str:
        """Return the `enc` from the JWE header.

        The `enc` header contains the identifier of the CEK encryption algorithm.

        Returns:
            the enc value

        Raises:
            AttributeError: if there is no enc header or it is not a string
        """
        enc = self.get_header("enc")
        if enc is None or not isinstance(enc, str):
            raise AttributeError("This JWE doesn't have a valid 'enc' header")
        return enc

    @classmethod
    def encrypt(
        cls,
        plaintext: bytes,
        jwk: Union[Jwk, Dict[str, Any]],
        *,
        enc: str,
        alg: Optional[str] = None,
        extra_headers: Optional[Dict[str, Any]] = None,
        cek: Optional[bytes] = None,
        iv: Optional[bytes] = None,
        epk: Optional[Jwk] = None,
    ) -> "JweCompact":
        """Encrypt an arbitrary plaintext and return the resulting JweCompact.

        Args:
          plaintext: the raw plaintext to encrypt
          jwk: the public or symmetric key to use
          enc: the CEK encryption algorithm to use
          alg: the Key Management algorithm to use, if there is no 'alg' header defined in the Jwk
          extra_headers: additional headers to include in the generated token
          cek: the CEK to force use, for algorithms relying on a random CEK. Leave `None` to have a safe value generated by `jwskate`.
          iv: the IV to force use. Leave `None` to have a safe value generated by `jwskate`.
          epk: the EPK to force use. Leave `None` to have a safe value generated by `jwskate`.

        Returns:
            the generated JweCompact instance
        """
        jwk = Jwk(jwk)
        extra_headers = extra_headers or {}
        cek_jwk, wrapped_cek, cek_headers = jwk.sender_key(
            enc=enc, alg=alg, cek=cek, epk=epk, **extra_headers
        )

        headers = dict(extra_headers, **cek_headers, alg=alg, enc=enc)
        aad = BinaPy.serialize_to("json", headers).to("b64u")

        ciphertext, iv, tag = cek_jwk.encrypt(plaintext, aad=aad, iv=iv, alg=enc)

        return cls.from_parts(
            headers=headers, cek=wrapped_cek, iv=iv, ciphertext=ciphertext, tag=tag
        )

    PBES2_ALGORITHMS: Mapping[str, Type[BasePbes2]] = {
        alg.name: alg
        for alg in [Pbes2_HS256_A128KW, Pbes2_HS384_A192KW, Pbes2_HS512_A256KW]
    }

    def unwrap_cek(
        self, jwk_or_password: Union[Jwk, Dict[str, Any], bytes, str]
    ) -> Jwk:
        """Unwrap the CEK from this JWE token using the provided key or password.

        Args:
          jwk_or_password: the decryption JWK or password

        Returns:
            the unwrapped CEK, as a SymmetricJwk
        """
        if isinstance(jwk_or_password, (bytes, str)):
            password = jwk_or_password
            return self.unwrap_cek_with_password(password)

        jwk = Jwk(jwk_or_password)
        cek = jwk.recipient_key(self.wrapped_cek, **self.headers)
        return cek

    def decrypt(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
    ) -> bytes:
        """Decrypts this Jwe payload using a JWK.

        Args:
          jwk: the decryption key

        Returns:
          bytes: the decrypted payload
        """
        cek_jwk = self.unwrap_cek(jwk)

        plaintext = cek_jwk.decrypt(
            ciphertext=self.ciphertext,
            iv=self.initialization_vector,
            tag=self.authentication_tag,
            aad=self.additional_authenticated_data,
            alg=self.enc,
        )
        return plaintext

    @classmethod
    def encrypt_with_password(
        cls,
        plaintext: bytes,
        password: Union[bytes, str],
        *,
        alg: str,
        enc: str,
        salt: Optional[bytes] = None,
        count: int = 2000,
        cek: Optional[bytes] = None,
        iv: Optional[bytes] = None,
    ) -> "JweCompact":
        """Encrypt a payload with a password and return the resulting JweCompact.

        This performs symmetric encryption using PBES2.

        Args:
          plaintext: the data to encrypt
          password: the password to use
          alg: the Key Management alg to use
          enc: the Payload Encryption alg to use
          salt: the salt to use. Leave `None` (default) to have `jwskate` generate a safe random value
          count: the number of PBES2 iterations (recommended minimum 1000)
          cek: the CEK to force use. Leave `None` (default) to have `jwskate` generate a safe random value
          iv: the IV to force use. Leave `None` (default) to have `jwskate` generate a safe random value

        Returns:
            the resulting JweCompact

        Raises:
            UnsupportedAlg: if the key management alg is not supported
            ValueError: if the `count` parameter is not a positive integer
        """
        keyalg = cls.PBES2_ALGORITHMS.get(alg)
        if keyalg is None:
            raise UnsupportedAlg(
                f"Unsupported password-based encryption algorithm '{alg}'"
            )

        if cek is None:
            cek_jwk = SymmetricJwk.generate_for_alg(enc)
            cek = cek_jwk.key
        else:
            cek_jwk = SymmetricJwk.from_bytes(cek)

        wrapper = keyalg(password)
        if salt is None:
            salt = wrapper.generate_salt()

        if count < 1:
            raise ValueError(
                "PBES2 iteration count must be a positive integer, with a minimum recommended value of 1000"
            )
        if count < 1000:
            warnings.warn("PBES2 iteration count should be > 1000")

        wrapped_cek = wrapper.wrap_key(cek, salt=salt, count=count)

        headers = dict(alg=alg, enc=enc, p2s=BinaPy(salt).to("b64u").ascii(), p2c=count)
        aad = BinaPy.serialize_to("json", headers).to("b64u")
        ciphertext, iv, tag = cek_jwk.encrypt(
            plaintext=plaintext, aad=aad, alg=enc, iv=iv
        )

        return cls.from_parts(
            headers=headers, cek=wrapped_cek, iv=iv, ciphertext=ciphertext, tag=tag
        )

    def unwrap_cek_with_password(self, password: Union[bytes, str]) -> Jwk:
        """Unwrap a CEK using a password. Works only for password-encrypted JWE Tokens.

        Args:
          password: the decryption password

        Returns:
            the CEK, as a SymmetricJwk instance

        Raises:
            UnsupportedAlg: if the token key management algorithm is not supported
            AttributeError: if the token misses the PBES2-related headers
        """
        keyalg = self.PBES2_ALGORITHMS.get(self.alg)
        if keyalg is None:
            raise UnsupportedAlg(
                f"Unsupported password-based encryption algorithm '{self.alg}'"
            )
        p2s = self.headers.get("p2s")
        if p2s is None:
            raise AttributeError("No 'p2s' in headers!")
        salt = BinaPy(p2s).decode_from("b64u")
        p2c = self.headers.get("p2c")
        if p2c is None:
            raise AttributeError("No 'p2c' in headers!")
        if not isinstance(p2c, int) or p2c < 1:
            raise AttributeError("Invalid value for p2c, must be a positive integer")
        wrapper = keyalg(password)
        cek = wrapper.unwrap_key(self.wrapped_cek, salt=salt, count=p2c)
        return SymmetricJwk.from_bytes(cek)

    def decrypt_with_password(self, password: Union[bytes, str]) -> bytes:
        """Decrypt the JWE token with a password. This only works for tokens encrypted with a password.

        Args:
          password: the password to use

        Returns:
            the unencrypted payload
        """
        cek_jwk = self.unwrap_cek_with_password(password)
        plaintext = cek_jwk.decrypt(
            ciphertext=self.ciphertext,
            iv=self.initialization_vector,
            tag=self.authentication_tag,
            aad=self.additional_authenticated_data,
            alg=self.enc,
        )
        return plaintext



import pytest

from jwskate import SymmetricJwk


@pytest.fixture(
    scope="module",
    params=(
        ("HS256", 256),
        ("HS384", 384),
        ("HS512", 512),
    ),
    ids=("256bits", "384bits", "512bits"),
)
def symmetric_jwk(request: pytest.FixtureRequest) -> SymmetricJwk:
    alg, min_key_size = request.param  # type: ignore
    kid = f"my_{alg}_jwk"
    jwk = SymmetricJwk.generate_for_alg(alg, kid=kid)
    assert jwk.kty == "oct"
    assert jwk.alg == alg
    assert jwk.kid == kid
    assert isinstance(jwk.k, str)
    assert jwk.key_size >= min_key_size

    with pytest.raises(ValueError):
        jwk.public_jwk()

    return jwk


def test_jwk_symmetric_sign(symmetric_jwk: SymmetricJwk) -> None:
    data = b"The true sign of intelligence is not knowledge but imagination."
    signature = symmetric_jwk.sign(data)
    assert symmetric_jwk.verify(data, signature)


def test_dir_alg(symmetric_jwk: SymmetricJwk) -> None:
    assert "dir" in symmetric_jwk.supported_key_management_algorithms()



"""This module implements AES-GCM based Key Management algorithms."""

from typing import Tuple

from binapy import BinaPy

from ..base import BaseKeyManagementAlg
from ..encryption.aesgcm import BaseAESGCM


class BaseAesGcmKeyWrap(BaseAESGCM, BaseKeyManagementAlg):
    """Base class for AES-GCM Key wrapping algorithms."""

    key_size: int
    """Required key size, in bits."""
    tag_size: int = 16
    """Authentication tag size, in bits."""
    iv_size: int = 96
    """Initialisation Vector size, in bits."""

    def wrap_key(self, plainkey: bytes, *, iv: bytes) -> Tuple[BinaPy, BinaPy]:
        """Wrap a key using the given Initialisation Vector (`iv`).

        Args:
          plainkey: the key to wrap
          iv: the Initialisation Vector to use

        Returns:
          a tuple (wrapped_key, authentication_tag)
        """
        return self.encrypt(plainkey, iv=iv, aad=b"")

    def unwrap_key(self, cipherkey: bytes, *, tag: bytes, iv: bytes) -> BinaPy:
        """Unwrap a key and authenticates it with the authentication `tag`, using the given Initialisation Vector (`iv`).

        Args:
          cipherkey: the wrapped key
          tag: the authentication tag
          iv: the Initialisation Vector

        Returns:
          the unwrapped key.
        """
        return self.decrypt(cipherkey, auth_tag=tag, iv=iv, aad=b"")


class A128GCMKW(BaseAesGcmKeyWrap):
    """Key wrapping with AES GCM using 128-bit key."""

    name = "A128GCMKW"
    description = __doc__
    key_size = 128


class A192GCMKW(BaseAesGcmKeyWrap):
    """Key wrapping with AES GCM using 192-bit key."""

    name = "A192GCMKW"
    description = __doc__
    key_size = 192


class A256GCMKW(BaseAesGcmKeyWrap):
    """Key wrapping with AES GCM using 256-bit key."""

    name = "A256GCMKW"
    description = __doc__
    key_size = 256



import pytest
from binapy import BinaPy

from jwskate import P_521, ECJwk, Jwk, JwsCompact, OKPJwk, RSAJwk, SymmetricJwk
from jwskate.jws.json import JwsJsonFlat, JwsJsonGeneral


def test_jws_compact(private_jwk: Jwk) -> None:
    jws = JwsCompact.sign(payload=b"Hello World!", jwk=private_jwk, alg="RS256")
    assert (
        str(jws)
        == "eyJhbGciOiJSUzI1NiIsImtpZCI6IkpXSy1BQkNEIn0.SGVsbG8gV29ybGQh.1eucS9ZaTnAJyfVNhxLJ_phFN1rexm0l-nIXWBjUImdS29z55BuxH6NjGpltSXKrgYxYQxqGCsGIxlSVoIEhKVdhE1Vd9NPJRyw7I4zBRdwVvcqMRODMqDxCiqbDQ_5bI5jAqFEJAFCXZo2T4ixlxs-2eXtmSEp6vX51Tg1pvicM5_YrKfS8Jn3lt9xW5RaNKUJ94KVLlov_IncFsh2bg5jdo1SEoUxlB2II0JdlfCsgHohJd58eWjFToeNtH1eiXGeZOHblMLz5a5AhY8jY3C424-tggj6BK6fwpedddFD3mtFFTNw6KT-2EgTeOlEA09pQqW5hosCj2duAlR-FQQ"
    )
    assert jws.verify_signature(private_jwk, alg="RS256")


EC_P521_PRIVATE_KEY = {
    "kty": "EC",
    "kid": "bilbo.baggins@hobbiton.example",
    "use": "sig",
    "crv": "P-521",
    "x": "AHKZLLOsCOzz5cY97ewNUajB957y-C-U88c3v13nmGZx6sYl_oJXu9"
    "A5RkTKqjqvjyekWF-7ytDyRXYgCF5cj0Kt",
    "y": "AdymlHvOiLxXkEhayXQnNCvDX4h9htZaCJN34kfmC6pV5OhQHiraVy"
    "SsUdaQkAgDPrwQrJmbnX9cwlGfP-HqHZR1",
    "d": "AAhRON2r9cqXX1hg-RoI6R1tX5p2rUAYdmpHZoC1XNM56KtscrX6zb"
    "KipQrCW9CGZH3T4ubpnoTKLDYJ_fF3_rJt",
}

RSA_PRIVATE_KEY = {
    "kty": "RSA",
    "kid": "bilbo.baggins@hobbiton.example",
    "use": "sig",
    "n": "n4EPtAOCc9AlkeQHPzHStgAbgs7bTZLwUBZdR8_KuKPEHLd4rHVTeT"
    "-O-XV2jRojdNhxJWTDvNd7nqQ0VEiZQHz_AJmSCpMaJMRBSFKrKb2wqV"
    "wGU_NsYOYL-QtiWN2lbzcEe6XC0dApr5ydQLrHqkHHig3RBordaZ6Aj-"
    "oBHqFEHYpPe7Tpe-OfVfHd1E6cS6M1FZcD1NNLYD5lFHpPI9bTwJlsde"
    "3uhGqC0ZCuEHg8lhzwOHrtIQbS0FVbb9k3-tVTU4fg_3L_vniUFAKwuC"
    "LqKnS2BYwdq_mzSnbLY7h_qixoR7jig3__kRhuaxwUkRz5iaiQkqgc5g"
    "HdrNP5zw",
    "e": "AQAB",
    "d": "bWUC9B-EFRIo8kpGfh0ZuyGPvMNKvYWNtB_ikiH9k20eT-O1q_I78e"
    "iZkpXxXQ0UTEs2LsNRS-8uJbvQ-A1irkwMSMkK1J3XTGgdrhCku9gRld"
    "Y7sNA_AKZGh-Q661_42rINLRCe8W-nZ34ui_qOfkLnK9QWDDqpaIsA-b"
    "MwWWSDFu2MUBYwkHTMEzLYGqOe04noqeq1hExBTHBOBdkMXiuFhUq1BU"
    "6l-DqEiWxqg82sXt2h-LMnT3046AOYJoRioz75tSUQfGCshWTBnP5uDj"
    "d18kKhyv07lhfSJdrPdM5Plyl21hsFf4L_mHCuoFau7gdsPfHPxxjVOc"
    "OpBrQzwQ",
    "p": "3Slxg_DwTXJcb6095RoXygQCAZ5RnAvZlno1yhHtnUex_fp7AZ_9nR"
    "aO7HX_-SFfGQeutao2TDjDAWU4Vupk8rw9JR0AzZ0N2fvuIAmr_WCsmG"
    "peNqQnev1T7IyEsnh8UMt-n5CafhkikzhEsrmndH6LxOrvRJlsPp6Zv8"
    "bUq0k",
    "q": "uKE2dh-cTf6ERF4k4e_jy78GfPYUIaUyoSSJuBzp3Cubk3OCqs6grT"
    "8bR_cu0Dm1MZwWmtdqDyI95HrUeq3MP15vMMON8lHTeZu2lmKvwqW7an"
    "V5UzhM1iZ7z4yMkuUwFWoBvyY898EXvRD-hdqRxHlSqAZ192zB3pVFJ0"
    "s7pFc",
    "dp": "B8PVvXkvJrj2L-GYQ7v3y9r6Kw5g9SahXBwsWUzp19TVlgI-YV85q"
    "1NIb1rxQtD-IsXXR3-TanevuRPRt5OBOdiMGQp8pbt26gljYfKU_E9xn"
    "-RULHz0-ed9E9gXLKD4VGngpz-PfQ_q29pk5xWHoJp009Qf1HvChixRX"
    "59ehik",
    "dq": "CLDmDGduhylc9o7r84rEUVn7pzQ6PF83Y-iBZx5NT-TpnOZKF1pEr"
    "AMVeKzFEl41DlHHqqBLSM0W1sOFbwTxYWZDm6sI6og5iTbwQGIC3gnJK"
    "bi_7k_vJgGHwHxgPaX2PnvP-zyEkDERuf-ry4c_Z11Cq9AqC2yeL6kdK"
    "T1cYF8",
    "qi": "3PiqvXQN0zwMeE-sBvZgi289XP9XCQF3VWqPzMKnIgQp7_Tugo6-N"
    "ZBKCQsMf3HaEGBjTVJs_jcK8-TRXvaKe-7ZMaQj8VfBdYkssbu0NKDDh"
    "jJ-GtiseaDVWt7dcH0cfwxgFUHpQh7FoCrjFJ6h6ZEpMF6xmujs4qMpP"
    "z8aaI4",
}

OKP_ED25519_PRIVATE_KEY = {
    "kty": "OKP",
    "crv": "Ed25519",
    "d": "nWGxne_9WmC6hEr0kuwsxERJxWl7MmkZcDusAxyuf2A",
    "x": "11qYAYKxCrfVS_7TyWQHOg7hcvPapiMlrwIaaPcHURo",
}

SYMMETRIC_SIGNATURE_KEY = {
    "kty": "oct",
    "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037",
    "use": "sig",
    "alg": "HS256",
    "k": "hJtXIZ2uSN5kbQfbtTNWbpdmhkV8FJG-Onbc6mxCcYg",
}


@pytest.fixture(scope="session")
def ec_p521_private_jwk() -> Jwk:
    """This is the key from [https://datatracker.ietf.org/doc/html/rfc7520#section-3.2]."""
    jwk = Jwk(EC_P521_PRIVATE_KEY)
    assert isinstance(jwk, ECJwk)
    assert jwk.is_private
    assert jwk.kty == "EC"
    assert jwk.kid == "bilbo.baggins@hobbiton.example"
    assert jwk.use == "sig"
    assert jwk.curve == P_521
    assert (
        jwk.x_coordinate
        == 1536512509633812701046363966946458604394346818697394258819956002474017850080242973018354677969345705661882653180474980600600249393774872942916765721086083757
    )
    assert (
        jwk.y_coordinate
        == 6390841077912737365653019441492828723058308087943275742908661475903115398365573971688270351422921162967673593344809147696590449694148618103053589689371825269
    )
    assert (
        jwk.ecc_private_key
        == 111516411687364059110290785888309499157003632541675704481981366909658327544236421609398595734137119337357107861242367481230550517706674527460992238934209133
    )
    return jwk


@pytest.fixture(scope="session")
def ec_p256_private_jwk() -> Jwk:
    return ECJwk.generate("P-256")


@pytest.fixture(scope="session")
def ec_p384_private_jwk() -> Jwk:
    return Jwk.generate_for_kty("EC", crv="P-384")


@pytest.fixture(scope="session")
def rsa_private_jwk() -> Jwk:
    """This is the key from [https://datatracker.ietf.org/doc/html/rfc7520#section-3.4]."""
    jwk = Jwk(RSA_PRIVATE_KEY)
    assert isinstance(jwk, RSAJwk)
    assert jwk.is_private
    assert jwk.kty == "RSA"
    assert jwk.kid == "bilbo.baggins@hobbiton.example"
    assert jwk.use == "sig"
    assert (
        jwk.modulus
        == 20135533008613362683983973718862990570890949482783547491074937566048838943004157274484500282679051238967930814182837332509745335321694730867914487474360313056717004122048241683576190451001206594369003880452220552186311851010130037332999299892700953157894377718386086768938058299374235398748350321163975673243254998238224668780038242796491971359194173117243083075284176788910883569789455869367283514387223800602948314723218768921623931105285092074647944930960873919066358313244754717122611255711161319444897038896496343014060976689972635082113758993979473481511550731351324005908071126862383299605264835961097030990287
    )
    assert jwk.exponent == 65537
    assert (
        jwk.private_exponent
        == 13809785886921180797407749068700942981528089435771470964933339849531763931979658226246689941649114165877756904281400924998825599768673188627050679509247407590724566295036763464811978094688530662125180988122227281988347728446577123121956338123153675672384095849459356212372062261627372823030362241721140402534827920608159114397698659666564432536888691574732214992533122683199444713216582417478411683595580959548347164585150169575401765252679902902625147728179779840007171037544685861186766305472768461620252400817122383292377247352244817836750259386865218586191336825522033117854784997616980884291365145982447543137217
    )
    assert (
        jwk.first_prime_factor
        == 155305159528675998315587554014523516083078608902053750652196202749677048642358299363759977556059891120561885641903854876605754500853726315968216134808579359395711784936178882676585818087673577574114127693348589707935410407050296794012012146664933439273320539907415872853360575628122941817407358922423140657993
    )
    assert (
        jwk.second_prime_factor
        == 129651410614568017951696388521026752738348674639303464401419464668770635900253174384526773731502315063946718183900420330879494363511129117909409612613133053606973655487402983938837056343590378935691659259752059752852280269661044647292328198275965901381648329420292300429253481090448036576608977166562458575959
    )
    assert (
        jwk.first_factor_crt_exponent
        == 5452754506240497308759433019323719585992094222860760795439270374399299463296556858507586681668276080205271813815413736836746282351411023590382002932611213604063854267637328669893708400136364221707380683009522939987478513612504889192694812845812469959990700860994002219137017021229395442602718050574418216489
    )
    assert (
        jwk.second_factor_crt_exponent
        == 6103034953475782159405883067387392346274709877813314428160871698478072108156934906636939392429995910283894984471867749832357906395346648896814450690625379104589983172538233447538219918824119490007305320907809395266022012480039103665056876141436971569684073061958920103517814199063615863387684736238234656863
    )
    assert (
        jwk.first_crt_coefficient
        == 155171362656114787674005338316026300971092335901511555016027916093530558354739494408751451514346305783601055276983183518496184496218932854791325892306914322459078533545141895619742270863660955772717492371592287055310307433879513025000807527757356866299049054535097378049521944951691464480088870568286553270414
    )
    return jwk


@pytest.fixture(scope="module")
def okp_ed25519_signature_jwk() -> Jwk:
    jwk = Jwk(OKP_ED25519_PRIVATE_KEY)
    assert isinstance(jwk, OKPJwk)
    assert jwk.is_private
    assert jwk.kty == "OKP"
    assert (
        jwk.private_key.hex()
        == "9d61b19deffd5a60ba844af492ec2cc44449c5697b326919703bac031cae7f60"
    )
    assert (
        jwk.public_key.hex()
        == "d75a980182b10ab7d54bfed3c964073a0ee172f3daa62325af021a68f707511a"
    )
    assert jwk.thumbprint() == "kPrK_qmxVWaYVA9wwBF6Iuo3vVzz7TxHCTwXBygrS4k"
    return jwk


@pytest.fixture(scope="session")
def symmetric_signature_jwk() -> Jwk:
    """This is the key from [https://datatracker.ietf.org/doc/html/rfc7520#section-3.5]."""
    jwk = Jwk(SYMMETRIC_SIGNATURE_KEY)
    assert isinstance(jwk, SymmetricJwk)
    assert jwk.is_private
    assert jwk.kty == "oct"
    assert jwk.kid == "018c0ae5-4d9b-471b-bfd6-eef314bc7037"
    assert jwk.use == "sig"
    assert jwk.alg == "HS256"
    assert (
        jwk.key.hex()
        == "849b57219dae48de646d07dbb533566e976686457c1491be3a76dcea6c427188"
    )
    assert jwk.key_size == 32 * 8
    return jwk


@pytest.fixture(scope="module")
def signature_payload() -> bytes:
    """This is the payload from [https://datatracker.ietf.org/doc/html/rfc7520#section-4]."""
    return (
        "It’s a dangerous business, Frodo, going out your door. "
        "You step onto the road, and if you don't keep your feet, "
        "there’s no knowing where you might be swept off to."
    ).encode()


@pytest.fixture(
    params=[
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "HS256",
        "HS384",
        "HS512",
        "PS256",
        "PS384",
        "PS512",
        "EdDSA",
    ]
)
def signature_alg(request: pytest.FixtureRequest) -> str:
    return request.param  # type: ignore[attr-defined,no-any-return]


@pytest.fixture()
def signature_jwk(
    signature_alg: str,
    rsa_private_jwk: Jwk,
    ec_p256_private_jwk: Jwk,
    ec_p384_private_jwk: Jwk,
    ec_p521_private_jwk: Jwk,
    okp_ed25519_signature_jwk: Jwk,
    symmetric_signature_jwk: Jwk,
) -> Jwk:
    for key in (
        rsa_private_jwk,
        ec_p521_private_jwk,
        ec_p384_private_jwk,
        ec_p256_private_jwk,
        okp_ed25519_signature_jwk,
        symmetric_signature_jwk,
    ):
        if signature_alg in key.supported_signing_algorithms():
            return key

    pytest.skip(f"No key supports this signature alg: {signature_alg}")


@pytest.fixture()
def validation_jwk(signature_jwk: Jwk) -> Jwk:
    if isinstance(signature_jwk, SymmetricJwk):
        return signature_jwk
    public_jwk = signature_jwk.public_jwk()
    assert not public_jwk.is_private
    assert signature_jwk.kty == public_jwk.kty
    return public_jwk


@pytest.fixture()
def signed_jws_compact(
    signature_payload: bytes, signature_jwk: Jwk, signature_alg: str
) -> JwsCompact:
    jws = JwsCompact.sign(
        payload=signature_payload, jwk=signature_jwk, alg=signature_alg
    )
    assert isinstance(jws, JwsCompact)
    return jws


@pytest.fixture()
def signed_jws_json_flat(
    signature_payload: bytes, signature_jwk: Jwk, signature_alg: str
) -> JwsJsonFlat:
    jws = JwsJsonFlat.sign(
        payload=signature_payload, jwk=signature_jwk, alg=signature_alg
    )
    assert isinstance(jws, JwsJsonFlat)
    return jws


@pytest.fixture()
def signed_jws_json_general(
    signature_payload: bytes, signature_jwk: Jwk, signature_alg: str
) -> JwsJsonGeneral:
    jws = JwsJsonGeneral.sign(signature_payload, (signature_jwk, signature_alg))
    assert isinstance(jws, JwsJsonGeneral)
    return jws


def test_validate_signature(
    signed_jws_compact: JwsCompact, validation_jwk: Jwk, signature_alg: str
) -> None:
    assert signed_jws_compact.verify_signature(validation_jwk, alg=signature_alg)


def test_validate_signature_json_flat(
    signed_jws_json_flat: JwsJsonFlat, validation_jwk: Jwk, signature_alg: str
) -> None:
    assert signed_jws_json_flat.verify_signature(validation_jwk, alg=signature_alg)


def test_validate_signature_json_general(
    signed_jws_json_general: JwsJsonGeneral, validation_jwk: Jwk, signature_alg: str
) -> None:
    assert signed_jws_json_general.verify_signature(validation_jwk, alg=signature_alg)


def test_jws_format_transformation(
    signed_jws_compact: JwsCompact,
    signed_jws_json_flat: JwsJsonFlat,
    signed_jws_json_general: JwsJsonGeneral,
    signature_alg: str,
) -> None:
    # signature is not deterministic for those algs, so those comparisons will fail
    if signature_alg not in (
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512",
    ):
        assert signed_jws_compact.flat_json() == signed_jws_json_flat
        assert signed_jws_compact.general_json() == signed_jws_json_general
        assert signed_jws_json_flat.generalize() == signed_jws_json_general
        assert signed_jws_json_general.flatten() == signed_jws_json_flat
        assert signed_jws_json_flat.compact() == signed_jws_compact


def test_validate_signature_by_jwcrypto(
    signed_jws_compact: JwsCompact, validation_jwk: Jwk, signature_alg: str
) -> None:
    """This test verifies tokens generated by `jwskate` using another lib `jwcrypto`.

    Args:
        signed_jws_compact: the Jws signed by jwskate to verify
        validation_jwk: the Jwk containing the verification key
        signature_alg: the signature alg
    """
    import jwcrypto.jwk  # type: ignore[import]
    import jwcrypto.jws  # type: ignore[import]

    jwk = jwcrypto.jwk.JWK(**validation_jwk)
    jws = jwcrypto.jws.JWS()
    jws.deserialize(str(signed_jws_compact))
    jws.verify(jwk)


@pytest.fixture()
def jwcrypto_signed_jws(
    signature_payload: bytes, signature_jwk: Jwk, signature_alg: str
) -> str:
    """Sign a JWS using `jwcrypto`, to make sure it validates with `jwskate`.

    Args:
        signature_payload: the payload to sign
        signature_jwk: the key to use
        signature_alg: the alg to use

    Returns:
        a JWS token
    """
    import jwcrypto.jwk
    import jwcrypto.jws

    jwk = jwcrypto.jwk.JWK(**signature_jwk)
    jws = jwcrypto.jws.JWS(signature_payload)

    jws.add_signature(
        jwk,
        alg=signature_alg,
        protected=BinaPy.serialize_to("json", {"alg": signature_alg}).decode(),
    )
    token: str = jws.serialize(True)
    return token


def test_validate_signature_from_jwcrypto(
    jwcrypto_signed_jws: str, validation_jwk: Jwk, signature_alg: str
) -> None:
    """Check that `jwskate`validates tokens signed by `jwcrypto`.

    Args:
        jwcrypto_signed_jws: the JWS to validate
        validation_jwk: the public key to validate the signature
        signature_alg: the alg to use
    """
    assert JwsCompact(jwcrypto_signed_jws).verify_signature(
        validation_jwk, alg=signature_alg
    )



"""This module implements AES-CBC with HMAC-SHA based Encryption algorithms."""

from typing import Optional, Tuple

from binapy import BinaPy
from cryptography import exceptions
from cryptography.hazmat.primitives import ciphers, constant_time, hashes, hmac, padding
from cryptography.hazmat.primitives.ciphers import algorithms, modes

from ..base import BaseAESEncryptionAlg


class BaseAesCbcHmacSha2(BaseAESEncryptionAlg):
    """Implements the family of AES-CBC with HMAC-SHA encryption algorithms."""

    mac_key_size: int
    """Required key size for the Hash algorithm, in bits."""

    aes_key_size: int
    """Required key size for the AES algorithm, in bits."""

    iv_size: int = 128
    """Initialization Vector size for the AES algorithm, in bits."""

    hash_alg: hashes.HashAlgorithm
    """Hash algorithm to use."""

    def __init_subclass__(cls) -> None:
        """This automatically sets the total key size based on the MAC and AES key sizes."""
        cls.key_size = cls.mac_key_size + cls.aes_key_size

    def __init__(self, key: bytes) -> None:
        """Initialize this wrapper with the given key.

        Args:
            key: the key to use for encryption and decryption.
        """
        super().__init__(key)
        self.mac_key = self.key[: self.mac_key_size // 8]
        self.aes_key = self.key[self.mac_key_size // 8 :]
        self.padding = padding.PKCS7(algorithms.AES.block_size)

    def mac(
        self, ciphertext: bytes, *, iv: bytes, aad: Optional[bytes] = None
    ) -> BinaPy:
        """Produce a Message Authentication Code for the given `ciphertext`, `iv` and `aad`.

        Args:
          ciphertext: the ciphertext
          iv: the Initialization Vector
          aad: the Additional Authenticated data

        Returns:
          the resulting MAC.
        """
        if aad is None:
            aad = b""
        al = BinaPy.from_int(len(aad) * 8, length=8, byteorder="big", signed=False)
        hasher = hmac.HMAC(self.mac_key, self.hash_alg)
        for param in (aad, iv, ciphertext, al):
            hasher.update(param)
        digest = hasher.finalize()
        mac = digest[: self.tag_size]
        return BinaPy(mac)

    def encrypt(
        self, plaintext: bytes, *, iv: bytes, aad: Optional[bytes] = None
    ) -> Tuple[BinaPy, BinaPy]:
        """Encrypt and MAC the given `plaintext`, using the given Initialization Vector (`iv`) and optional Additional Authenticated Data (`aad`).

        Args:
          plaintext: the plain data to encrypt
          iv: the Initialization Vector
          aad: the Additional Authenticated Data, if any

        Returns:
          a tuple (encrypted_data, authentication_tag)
        """
        cipher = ciphers.Cipher(algorithms.AES(self.aes_key), modes.CBC(iv)).encryptor()
        padder = self.padding.padder()
        padded_text = padder.update(plaintext) + padder.finalize()
        ciphertext = cipher.update(padded_text) + cipher.finalize()
        mac = self.mac(ciphertext, iv=iv, aad=aad)
        return BinaPy(ciphertext), BinaPy(mac)

    def decrypt(
        self, ciphertext: bytes, *, iv: bytes, auth_tag: bytes, aad: Optional[bytes]
    ) -> BinaPy:
        """Decrypt and authenticate the given ciphertext with authentication tag (`ciphertext_with_tag`), as produced by `encrypt()`.

        Args:
          ciphertext: the ciphertext
          auth_tag: the authentication tag
          iv: the Initialization Vector
          aad: the Additional Authenticated Data, if any

        Returns:
          the decrypted data
        """
        mac = self.mac(ciphertext, iv=iv, aad=aad)
        if not constant_time.bytes_eq(mac, auth_tag):
            raise exceptions.InvalidSignature()

        cipher = ciphers.Cipher(algorithms.AES(self.aes_key), modes.CBC(iv)).decryptor()
        padded_text = cipher.update(ciphertext) + cipher.finalize()
        unpadder = self.padding.unpadder()
        return BinaPy(unpadder.update(padded_text) + unpadder.finalize())


class Aes128CbcHmacSha256(BaseAesCbcHmacSha2):
    """AES_128_CBC_HMAC_SHA_256."""

    name = "A128CBC-HS256"
    description = __doc__
    mac_key_size = 128
    aes_key_size = 128
    tag_size = 16
    hash_alg = hashes.SHA256()


class Aes192CbcHmacSha384(BaseAesCbcHmacSha2):
    """AES_192_CBC_HMAC_SHA_384."""

    name = "A192CBC-HS384"
    description = __doc__
    mac_key_size = 192
    aes_key_size = 192
    tag_size = 24
    hash_alg = hashes.SHA384()


class Aes256CbcHmacSha512(BaseAesCbcHmacSha2):
    """AES_256_CBC_HMAC_SHA_512."""

    name = "A256CBC-HS512"
    description = __doc__
    key_size = 512
    mac_key_size = 256
    aes_key_size = 256
    tag_size = 32
    hash_alg = hashes.SHA512()



"""This module contains classes that describe CFRG Elliptic Curve Diffie- Hellman algorithms as specified in RFC8037."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Tuple, Type, Union

# to support Python 3.7
try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore[misc]

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed448, ed25519, x448, x25519


class PublicKeyProtocol(Protocol):  # noqa
    def public_bytes(
        self,
        encoding: serialization.Encoding,
        format: serialization.PublicFormat,
    ) -> bytes:  # noqa
        ...


class PrivateKeyProtocol(Protocol):  # noqa
    def private_bytes(
        self,
        encoding: serialization.Encoding,
        format: serialization.PrivateFormat,
        encryption_algorithm: serialization.KeySerializationEncryption,
    ) -> bytes:  # noqa
        ...

    def public_key(self) -> PublicKeyProtocol:  # noqa
        ...

    @classmethod
    def generate(self) -> PrivateKeyProtocol:  # noqa
        ...


@dataclass
class OKPCurve:
    """Represent an Octet Key Pair (OKP) Curve."""

    name: str
    """Curve name as defined in [IANA JOSE](https://www.iana.org/assignments/jose/jose.xhtml#web-key-elliptic-curve).
    This name will appear in `alg` headers."""

    description: str
    """Curve description (human readable)."""

    cryptography_private_key_class: Type[Any]
    """`cryptography` private key class."""

    cryptography_public_key_class: Type[Any]
    """`cryptography` public key class."""

    use: str
    """Curve usage (`'sig'` or '`enc'`)."""

    instances: ClassVar[Dict[str, OKPCurve]] = {}
    """Registry of subclasses, in a {name: instance} mapping."""

    def __post_init__(self) -> None:
        """Automatically registers subclasses in the instance registry."""
        self.instances[self.name] = self

    def generate(self) -> Tuple[bytes, bytes]:
        """Generate a new private key on this curve.

        Returns:
            a tuple of `x` (public  part), and `d` (private part), as bytes
        """
        key = self.cryptography_private_key_class.generate()
        x = key.public_key().public_bytes(
            serialization.Encoding.Raw, serialization.PublicFormat.Raw
        )
        d = key.private_bytes(
            serialization.Encoding.Raw,
            serialization.PrivateFormat.Raw,
            serialization.NoEncryption(),
        )
        return x, d

    @classmethod
    def get_curve(cls, key: Union[PublicKeyProtocol, PrivateKeyProtocol]) -> OKPCurve:
        """Return the appropriate `OKPCurve` instance for a given `cryptography` private or public key.

        Args:
          key: `cryptography` private or public OKP key.

        Returns:
          OKPCurve: the appropriate `OKPCurve` for the given key

        Raises:
            NotImplementedError: if the required OKP curve is not supported
        """
        for c in cls.instances.values():
            if isinstance(
                key, (c.cryptography_private_key_class, c.cryptography_public_key_class)
            ):
                return c
        raise NotImplementedError(f"Unsupported OKP key {type(key)}")


Ed25519 = OKPCurve(
    name="Ed25519",
    description="Ed25519 signature algorithm key pairs",
    cryptography_private_key_class=ed25519.Ed25519PrivateKey,
    cryptography_public_key_class=ed25519.Ed25519PublicKey,
    use="sig",
)
"""Ed25519 curve."""

Ed448 = OKPCurve(
    name="Ed448",
    description="Ed448 signature algorithm key pairs",
    cryptography_private_key_class=ed448.Ed448PrivateKey,
    cryptography_public_key_class=ed448.Ed448PublicKey,
    use="sig",
)
"""Ed448 curve."""

X25519 = OKPCurve(
    name="X25519",
    description="X25519 function key pairs",
    cryptography_private_key_class=x25519.X25519PrivateKey,
    cryptography_public_key_class=x25519.X25519PublicKey,
    use="enc",
)
"""X25519 curve."""

X448 = OKPCurve(
    name="X448",
    description="X448 function key pairs",
    cryptography_private_key_class=x448.X448PrivateKey,
    cryptography_public_key_class=x448.X448PublicKey,
    use="enc",
)
"""X448 curve."""



"""This module implement Elliptic Curve signature algorithms."""

from binapy import BinaPy
from cryptography import exceptions
from cryptography.hazmat.primitives import asymmetric, hashes
from cryptography.hazmat.primitives.asymmetric import ec

from ..base import BaseAsymmetricAlg, BaseSignatureAlg
from ..ec import P_256, P_384, P_521, EllipticCurve, secp256k1


class BaseECSignatureAlg(
    BaseAsymmetricAlg[ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey],
    BaseSignatureAlg,
):
    """Base class for Elliptic Curve signature algorithms."""

    curve: EllipticCurve
    hashing_alg: hashes.HashAlgorithm
    public_key_class = ec.EllipticCurvePublicKey
    private_key_class = ec.EllipticCurvePrivateKey

    def sign(self, data: bytes) -> BinaPy:  # noqa: D102
        with self.private_key_required() as key:
            dss_sig = key.sign(data, ec.ECDSA(self.hashing_alg))
            r, s = asymmetric.utils.decode_dss_signature(dss_sig)
            return BinaPy.from_int(r, self.curve.coordinate_size) + BinaPy.from_int(
                s, self.curve.coordinate_size
            )

    def verify(self, data: bytes, signature: bytes) -> bool:  # noqa: D102
        with self.public_key_required() as key:
            if len(signature) != self.curve.coordinate_size * 2:
                raise ValueError(
                    f"Invalid signature length {len(signature)} bytes, expected {self.curve.coordinate_size * 2} bytes"
                )

            r_bytes, s_bytes = (
                signature[: self.curve.coordinate_size],
                signature[self.curve.coordinate_size :],
            )
            r = int.from_bytes(r_bytes, "big", signed=False)
            s = int.from_bytes(s_bytes, "big", signed=False)
            dss_signature = asymmetric.utils.encode_dss_signature(r, s)

            try:
                key.verify(
                    dss_signature,
                    data,
                    ec.ECDSA(self.hashing_alg),
                )
                return True
            except exceptions.InvalidSignature:
                return False


class ES256(BaseECSignatureAlg):  # noqa: D415
    """ECDSA using P-256 and SHA-256"""

    name = "ES256"
    description = __doc__
    curve = P_256
    hashing_alg = hashes.SHA256()


class ES384(BaseECSignatureAlg):  # noqa: D415
    """ECDSA using P-384 and SHA-384"""

    name = "ES384"
    description = __doc__
    curve = P_384
    hashing_alg = hashes.SHA384()


class ES512(BaseECSignatureAlg):  # noqa: D415
    """ECDSA using P-521 and SHA-512"""

    name = "ES512"
    description = __doc__
    curve = P_521
    hashing_alg = hashes.SHA512()


class ES256K(BaseECSignatureAlg):  # noqa: D415
    """ECDSA using secp256k1 and SHA-256"""

    name = "ES256k"
    description = __doc__
    curve = secp256k1
    hashing_alg = hashes.SHA256()



"""This module implements the Json Web Algorithms as defined in RFC7518.

Each algorithm is represented as a wrapper around a symmetric or
asymmetric key, and exposes the cryptographic operations as methods. The
cryptographic operations themselves are delegated to `cryptography`.
"""

from .base import (
    BaseAESEncryptionAlg,
    BaseAlg,
    BaseAsymmetricAlg,
    BaseKeyManagementAlg,
    BaseSignatureAlg,
    BaseSymmetricAlg,
    PrivateKeyRequired,
    PublicKeyRequired,
)
from .ec import P_256, P_384, P_521, EllipticCurve, secp256k1
from .encryption import (
    A128GCM,
    A192GCM,
    A256GCM,
    Aes128CbcHmacSha256,
    Aes192CbcHmacSha384,
    Aes256CbcHmacSha512,
)
from .key_mgmt import (
    A128GCMKW,
    A128KW,
    A192GCMKW,
    A192KW,
    A256GCMKW,
    A256KW,
    BaseAesGcmKeyWrap,
    BaseAesKeyWrap,
    BaseEcdhEs_AesKw,
    BasePbes2,
    BaseRsaKeyWrap,
    DirectKeyUse,
    EcdhEs,
    EcdhEs_A128KW,
    EcdhEs_A192KW,
    EcdhEs_A256KW,
    Pbes2_HS256_A128KW,
    Pbes2_HS384_A192KW,
    Pbes2_HS512_A256KW,
    RsaEsOaep,
    RsaEsOaepSha256,
    RsaEsOaepSha384,
    RsaEsOaepSha512,
    RsaEsPcks1v1_5,
)
from .okp import X448, X25519, Ed448, Ed25519, OKPCurve
from .signature import (
    ES256,
    ES256K,
    ES384,
    ES512,
    HS256,
    HS384,
    HS512,
    PS256,
    PS384,
    PS512,
    RS256,
    RS384,
    RS512,
    EdDsa,
)

__all__ = [
    "BaseAlg",
    "BaseAsymmetricAlg",
    "BaseSymmetricAlg",
    "BaseAESEncryptionAlg",
    "BaseKeyManagementAlg",
    "BaseSignatureAlg",
    "P_256",
    "P_384",
    "P_521",
    "EllipticCurve",
    "secp256k1",
    "Aes128CbcHmacSha256",
    "Aes192CbcHmacSha384",
    "Aes256CbcHmacSha512",
    "A128GCMKW",
    "A192GCMKW",
    "A256GCMKW",
    "BaseAesGcmKeyWrap",
    "A128GCM",
    "A192GCM",
    "A256GCM",
    "BaseAesKeyWrap",
    "A128KW",
    "A192KW",
    "A256KW",
    "DirectKeyUse",
    "EcdhEs",
    "BaseEcdhEs_AesKw",
    "EcdhEs_A128KW",
    "EcdhEs_A192KW",
    "EcdhEs_A256KW",
    "BasePbes2",
    "Pbes2_HS256_A128KW",
    "Pbes2_HS384_A192KW",
    "Pbes2_HS512_A256KW",
    "BaseRsaKeyWrap",
    "RsaEsOaep",
    "RsaEsOaepSha256",
    "RsaEsOaepSha384",
    "RsaEsOaepSha512",
    "RsaEsPcks1v1_5",
    "HS256",
    "HS384",
    "HS512",
    "EdDsa",
    "ES256",
    "ES256K",
    "ES384",
    "ES512",
    "PS256",
    "PS384",
    "PS512",
    "RS256",
    "RS384",
    "RS512",
    "OKPCurve",
    "Ed25519",
    "Ed448",
    "X448",
    "X25519",
    "PrivateKeyRequired",
    "PublicKeyRequired",
]



"""Main module for `jwskate`.

The `jwskate` module implements the various Json Web Crypto-related
standards: JWA, JWK, JWKS, JWE, JWT. Each standard has its own submodule, but
for convenience, you can import any class or component directly from the root
`jwskate` module.

`jwskate` doesn't implement any actual cryptographic operation, it just
provides a set of convenient wrappers around the `cryptography` module.
"""

__author__ = """Guillaume Pujol"""
__email__ = "guill.p.linux@gmail.com"

from .enums import EncryptionAlgs, KeyManagementAlgs, SignatureAlgs
from .jwa import (
    A128GCM,
    A128GCMKW,
    A128KW,
    A192GCM,
    A192GCMKW,
    A192KW,
    A256GCM,
    A256GCMKW,
    A256KW,
    ES256,
    ES256K,
    ES384,
    ES512,
    HS256,
    HS384,
    HS512,
    P_256,
    P_384,
    P_521,
    PS256,
    PS384,
    PS512,
    RS256,
    RS384,
    RS512,
    X448,
    X25519,
    Aes128CbcHmacSha256,
    Aes192CbcHmacSha384,
    Aes256CbcHmacSha512,
    BaseAESEncryptionAlg,
    BaseAlg,
    BaseAsymmetricAlg,
    BaseEcdhEs_AesKw,
    BaseKeyManagementAlg,
    BasePbes2,
    BaseRsaKeyWrap,
    BaseSignatureAlg,
    BaseSymmetricAlg,
    DirectKeyUse,
    EcdhEs,
    EcdhEs_A128KW,
    EcdhEs_A192KW,
    EcdhEs_A256KW,
    Ed448,
    Ed25519,
    EdDsa,
    EllipticCurve,
    OKPCurve,
    Pbes2_HS256_A128KW,
    Pbes2_HS384_A192KW,
    Pbes2_HS512_A256KW,
    PrivateKeyRequired,
    PublicKeyRequired,
    RsaEsOaep,
    RsaEsOaepSha256,
    RsaEsOaepSha384,
    RsaEsOaepSha512,
    RsaEsPcks1v1_5,
    secp256k1,
)
from .jwe import InvalidJwe, JweCompact
from .jwk import (
    ECJwk,
    ExpectedAlgRequired,
    InvalidJwk,
    Jwk,
    JwkSet,
    OKPJwk,
    RSAJwk,
    SymmetricJwk,
    UnsupportedAlg,
    UnsupportedEllipticCurve,
    UnsupportedOKPCurve,
)
from .jws import InvalidJws, JwsCompact
from .jwt import (
    EncryptedJwt,
    ExpiredJwt,
    InvalidClaim,
    InvalidJwt,
    InvalidSignature,
    Jwt,
    JwtSigner,
    SignedJwt,
)



"""This module implements direct use of a shared symmetric key as Key Management algorithm."""

from typing import Type

from binapy import BinaPy

from ..base import BaseKeyManagementAlg, BaseSymmetricAlg


class DirectKeyUse(BaseKeyManagementAlg, BaseSymmetricAlg):
    """Direct use of a shared symmetric key as the CEK."""

    name = "dir"
    description = __doc__

    def direct_key(self, aesalg: Type[BaseSymmetricAlg]) -> BinaPy:
        """Check that the current key is appropriate for a given alg and return that same key.

        Args:
          aesalg: the AES encryption alg to use

        Returns:
          the current configured key, as-is
        """
        aesalg.check_key(self.key)
        return BinaPy(self.key)



from datetime import datetime, timezone

import pytest

from jwskate import (
    ExpectedAlgRequired,
    ExpiredJwt,
    InvalidClaim,
    InvalidJwt,
    InvalidSignature,
    Jwk,
    Jwt,
    JwtSigner,
    SignatureAlgs,
    SignedJwt,
    SymmetricJwk,
    UnsupportedAlg,
)


def test_jwt() -> None:
    jwt = Jwt(
        "eyJhbGciOiJSUzI1NiIsImtpZCI6Im15X2tleSJ9.eyJhY3IiOiIyIiwiYW1yIjpbInB3ZCIsIm90cCJdLCJhdWQiOiJjbGllbnRfaWQiLCJhdXRoX3RpbWUiOjE2MjkyMDQ1NjAsImV4cCI6MTYyOTIwNDYyMCwiaWF0IjoxNjI5MjA0NTYwLCJpc3MiOiJodHRwczovL215YXMubG9jYWwiLCJub25jZSI6Im5vbmNlIiwic3ViIjoiMTIzNDU2In0.wUfjMyjlOSdvbFGFP8O8wGcNBK7akeyOUBMvYcNZclFUtokOyxhLUPxmo1THo1DV1BHUVd6AWfeKUnyTxl_8-G3E_a9u5wJfDyfghPDhCmfkYARvqQnnV_3aIbfTfUBC4f0bHr08d_q0fED88RLu77wESIPCVqQYy2bk4FLucc63yGBvaCskqzthZ85DbBJYWLlR8qBUk_NA8bWATYEtjwTrxoZe-uA-vB6NwUv1h8DKRsDF-9HSVHeWXXAeoG9UW7zgxoY3KbDIVzemvGzs2R9OgDBRRafBBVeAkDV6CdbdMNJDmHzcjase5jX6LE-3YCy7c7AMM1uWRCnK3f-azA"
    )

    assert isinstance(jwt, SignedJwt)
    assert jwt.headers == {"alg": "RS256", "kid": "my_key"}
    assert jwt.claims == {
        "acr": "2",
        "amr": ["pwd", "otp"],
        "aud": "client_id",
        "auth_time": 1629204560,
        "exp": 1629204620,
        "iat": 1629204560,
        "iss": "https://myas.local",
        "nonce": "nonce",
        "sub": "123456",
    }
    assert jwt.is_expired()
    assert jwt.sub == "123456"
    assert jwt.subject == "123456"
    assert jwt.audiences == ["client_id"]
    assert jwt.nonce == "nonce"
    assert jwt.amr == ["pwd", "otp"]
    assert jwt.exp == 1629204620
    assert jwt.expires_at == datetime.fromtimestamp(1629204620, tz=timezone.utc)
    assert jwt.issued_at == datetime.fromtimestamp(1629204560, tz=timezone.utc)
    assert jwt.nonce == jwt["nonce"]
    jwt.validate(
        jwk=Jwk(
            {
                "kty": "RSA",
                "alg": "RS256",
                "kid": "my_key",
                "n": "2m4QVSHdUo2DFSbGY24cJbxE10KbgdkSCtm0YZ1q0Zmna8pJg8YhaWCJHV7D5AxQ_L1b1PK0jsdpGYWc5-Pys0FB2hyABGPxXIdg1mjxn6geHLpWzsA3MHD29oqfl0Rt7g6AFc5St3lBgJCyWtci6QYBmBkX9oIMOx9pgv4BaT6y1DdrNh27-oSMXZ0a58KwnC6jbCpdA3V3Eume-Be1Tx9lJN3j6S8ydT7CGY1Xd-sc3oB8pXfkr1_EYf0Sgb9EwOJfqlNK_kVjT3GZ-1JJMKJ6zkU7H0yXe2SKXAzfayvJaIcYrk-sYwmf-u7yioOLLvjlGjysN7SOSM8socACcw",
                "e": "AQAB",
                "d": "RldleRTzwi8CRKB9CO4fsGNFxBCWJaWy8r2TIlBgYulZihPVwtLeVaIZ5dRrvxfcSNfuJ9CVJtm-1dI6ak71DJb6TvQYodFRm9uY6tNW5HRuZg_3_pLV8wqd7V1M8Zi-0gfnZZ5Q8vbgijeOyEQ54NLnVoTWO7M7nxqJjv6fk7Vd1vd6Gy8jI_soA6AMFCSAF-Vab07jGklBaLyow_TdczYufQ1737RNsFra2l43esAKeavxxkr7Js6OpgUkrXPEOc19GAwJLDdfkZ6yJLR8poWwX_OD-Opmvqmq6BT0s0mAyjBKZUxTGJuD3hm6mKOxXrbJOKY_UXRN7EAuH6U0gQ",
                "p": "9WQs9id-xB2AhrpHgyt4nfljXFXjaDqRHzUydw15HAOoSZzYMZJW-GT8g2hB3oH3EsSCuMh70eiE1ohTLeipYdJ-s7Gy5qTH5-CblT_OfLXxi2hIumdTx53w-AtDEWl2PRt_qGHZ0B83NjVU2fo96kp9bgJWYh_iWWtSJyabXbM",
                "q": "499_fCUhh5zL-3a4WGENy_yrsAa5C1sylZUtokyJNYBz68kWRFHFsArXnwZifBD_GWBgJQtldsouqvvPxzAlHQB9kfhxaRbaugwVePSjgHYmhd-NhAySq7rBURvRquAxJmoBmN2lS54YyN_X-VAKgfHDNsN7f7LIw9ISrLeR6EE",
                "dp": "Cfxwo_fJfduhfloYTOs49lzOwVQxc-1mOHnmuteOhShU8eHzHllRNryNVh-pBpANaPMcSr7F4y3uMfjMQcMFGZkCVPe3SxGLnRET48f79DFHSiANTaCk1SvFQaLbsNq02BnFYSnSPlj22zriYBiB6oXrgs2PjGC1ymPGrRcyHWc",
                "dq": "hL-4AfeTn_AtORJBdGMd6X8J-eMAu-fmARRF4G3b5Qou_eZIjYZhtxup31-V0hcItZzahdoswtYn9734nl6i0FFv1bC5SPJie838WFmUQosSCB1i0NGORHLombquG3C90VYiFg7Rc8rnP2Z_6CLD7E2OXwHkmVDq-oEQFgRfAME",
                "qi": "riPJlv9XNjdheryQWGr7Rhlvp9rxeNyWfVzj3y_IGh3tpe--Cd6-1GUrF00HLTTc-5iKVIa-FWOeMPTYc2_Uldi_0qWlrKjM5teIpUlDJbz7Ha-bfed9-eTbG8cI5F57KdDjbjB8YgqWYKz4YPMwqZFbWxZi4W_X79Bs3htXcXA",
            }
        ),
        issuer="https://myas.local",
        audience="client_id",
        check_exp=False,
    )

    with pytest.raises(InvalidSignature):
        jwt.validate(Jwk.generate_for_kty("RSA"), alg="RS256")


def test_unprotected() -> None:
    jwt = Jwt.unprotected({"foo": "bar"})
    assert jwt == "eyJhbGciOiJub25lIn0.eyJmb28iOiJiYXIifQ."


def test_jwt_signer(issuer: str, private_jwk: Jwk) -> None:
    signer = JwtSigner(issuer, private_jwk)
    now = datetime.now(timezone.utc)
    jwt = signer.sign(subject="some_id", audience="some_audience")
    assert isinstance(jwt, Jwt)
    assert jwt.subject == "some_id"
    assert jwt.audiences == ["some_audience"]
    assert jwt.iat == pytest.approx(now.timestamp())
    assert jwt.expires_at is not None
    assert jwt.expires_at > now


def test_invalid_signed_jwt() -> None:
    with pytest.raises(InvalidJwt):
        SignedJwt("garbage")
    with pytest.raises(InvalidJwt):
        SignedJwt("garbage!.foo!.bar!")
    with pytest.raises(InvalidJwt):
        SignedJwt("eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.garbage!.garbage!")
    with pytest.raises(InvalidJwt):
        SignedJwt(
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ.garbage!"
        )


def test_empty_jwt(private_jwk: Jwk) -> None:
    jwt = Jwt.sign({}, private_jwk)
    assert jwt.is_expired() is None
    assert jwt.expires_at is None
    assert jwt.not_before is None
    assert jwt.issuer is None
    assert jwt.audiences is None
    assert jwt.subject is None
    assert jwt.jwt_token_id is None
    assert jwt.kid == private_jwk.kid
    assert jwt.alg == private_jwk.alg

    with pytest.raises(AttributeError):
        jwt.foo

    with pytest.raises(KeyError):
        jwt["foo"]

    assert (
        str(jwt)
        == "eyJhbGciOiJSUzI1NiIsImtpZCI6IkpXSy1BQkNEIn0.e30.iQ6eyP9QrNDQVfKYwHpIWvnTBb0SXNEItDxMQ0VmsB5CA5FaRYtGvj01VjoUAHEegGvwFI4YI35MDY_-DUR3UIXqxVMOe9Hk2hb9paTjJLpDIa7Ml6LKDh9-4xmAAPjZcra4IYrpDux8ohg0LUzgxHn0xnKDuxKmlh9shCyNWEOfN_i_JX4v8aSD-zDBteftrY9GzHU4Y0mlvlm4FaAwafPXovZilb_dTgTBkiFLmXY0y5ESurhQ4LrQqLzBz45lSONLElil5OQu0ySrrC72CBKp-HjdpCsLyG9F9p_X-1r9mmqlN38zIcgwkbhek04ieX-rumyzvHLO_UzttxDQOg"
    )
    assert bytes(jwt) == str(jwt).encode()
    assert jwt.signed_part == b"eyJhbGciOiJSUzI1NiIsImtpZCI6IkpXSy1BQkNEIn0.e30"

    jwt.validate(jwk=private_jwk, check_exp=False)

    with pytest.raises(InvalidClaim):
        jwt.validate(jwk=private_jwk)


def test_validate() -> None:
    jwt = SignedJwt(
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaXNzIjoiaHR0cHM6Ly9pc3N1ZXIubG9jYWwiLCJhdWQiOiJodHRwczovL2F1ZGllbmNlLmxvY2FsIiwiZXhwIjoyMDAwMDAwMDAwLCJpYXQiOjE1MTYyMzkwMjIsImNsYWltMSI6IkkgaGF2ZSBhIDEifQ.bl5iNgXfkbmgDXItaUx7_1lUMNtOffihsShVP8MeE1g"
    )
    jwk = SymmetricJwk.from_bytes("your-256-bit-secret")

    jwt.validate(
        jwk,
        iss="https://issuer.local",
        audience="https://audience.local",
        sub="1234567890",
        name="John Doe",
        algs=SignatureAlgs.ALL_SYMMETRIC,
        claim1=lambda value: "1" in value,
    )

    with pytest.raises(ExpectedAlgRequired):
        # expected algs must be provided, unless jwk has an 'alg' parameter
        jwt.validate(jwk)

    with pytest.raises(UnsupportedAlg):
        # unsupported alg
        jwt.validate(jwk, alg="foobar")

    with pytest.raises(UnsupportedAlg):
        # unsupported alg
        jwt.validate(jwk, algs=SignatureAlgs.ALL_ASYMMETRIC)

    with pytest.raises(InvalidClaim):
        jwt.validate(jwk, sub="invalid_sub", algs=SignatureAlgs.ALL_SYMMETRIC)

    with pytest.raises(InvalidClaim):
        jwt.validate(jwk, issuer="invalid_iss", algs=SignatureAlgs.ALL_SYMMETRIC)

    with pytest.raises(InvalidClaim):
        jwt.validate(
            jwk, algs=SignatureAlgs.ALL_SYMMETRIC, claim1=lambda value: "2" in value
        )

    with pytest.raises(ExpiredJwt):
        SignedJwt(
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaXNzIjoiaHR0cHM6Ly9pc3N1ZXIubG9jYWwiLCJhdWQiOiJodHRwczovL2F1ZGllbmNlLmxvY2FsIiwiZXhwIjoxNTE2MjM5MDIyLCJpYXQiOjE1MTYyMzkwMjIsImNsYWltMSI6IkkgaGF2ZSBhIDEifQ.k4qhY14C0sJYTaUiAIc2kkybmaIxaUMkirIkln10SG4"
        ).validate(jwk, algs=SignatureAlgs.ALL_SYMMETRIC)



"""This modules contains classes and utilities to generate and validate signed JWT."""

from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Iterable, List, Optional, Union

from backports.cached_property import cached_property
from binapy import BinaPy

from jwskate import Jwk

from .base import InvalidJwt, Jwt


class ExpiredJwt(ValueError):
    """Raised when trying to validate an expired JWT token."""


class InvalidSignature(ValueError):
    """Raised when trying to validate a JWT with an invalid signature."""


class InvalidClaim(ValueError):
    """Raised when trying to validate a JWT with unexpected claims."""


class SignedJwt(Jwt):
    """Represent a Signed Json Web Token (JWT), as defined in RFC7519.

    A signed JWT contains a JSON object as payload, which represents claims.

    To sign a JWT, use [Jwt.sign][jwskate.jwt.Jwt.sign].
    """

    def __init__(self, value: Union[bytes, str]) -> None:
        """Initialize a `SignedJwt`, from its compact serialized value.

        Args:
            value: the token value.
        """
        super().__init__(value)

        if self.value.count(b".") != 2:
            raise InvalidJwt(
                "A JWT must contain a header, a payload and a signature, separated by dots",
                value,
            )

        header, payload, signature = self.value.split(b".")
        try:
            self.headers = BinaPy(header).decode_from("b64u").parse_from("json")
        except ValueError:
            raise InvalidJwt(
                "Invalid JWT header: it must be a Base64URL-encoded JSON object"
            )

        try:
            self.claims = BinaPy(payload).decode_from("b64u").parse_from("json")
        except ValueError:
            raise InvalidJwt(
                "Invalid JWT payload: it must be a Base64URL-encoded JSON object"
            )

        try:
            self.signature = BinaPy(signature).decode_from("b64u")
        except ValueError:
            raise InvalidJwt(
                "Invalid JWT signature: it must be a Base64URL-encoded binary data (bytes)"
            )

    @cached_property
    def signed_part(self) -> bytes:
        """Return the actual signed data from this token.

        The signed part is composed of the header and payload, encoded in Base64-Url, joined by a dot.

        Returns:
          the signed part as bytes
        """
        return b".".join(self.value.split(b".", 2)[:2])

    def verify_signature(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
    ) -> bool:
        """Verify this JWT signature using a given key and algorithm(s).

        Args:
          jwk: the private Jwk to use to verify the signature
          alg: the alg to use to verify the signature, if only 1 is allowed
          algs: the allowed signature algs, if there are several

        Returns:
            `True` if the token signature is verified, `False` otherwise
        """
        jwk = Jwk(jwk)

        return jwk.verify(
            data=self.signed_part, signature=self.signature, alg=alg, algs=algs
        )

    def is_expired(self, leeway: int = 0) -> Optional[bool]:
        """Check if this token is expired, based on its `exp` claim.

        Args:
            leeway: additional number of seconds for leeway.

        Returns:
            `True` if the token is expired, `False` if it's not, `None` if there is no `exp` claim.
        """
        exp = self.expires_at
        if exp is None:
            return None
        return exp < (datetime.now(timezone.utc) + timedelta(seconds=leeway))

    @cached_property
    def expires_at(self) -> Optional[datetime]:
        """Get the "Expires At" (exp) date from this token.

        Returns:
          a `datetime` initialized from the `exp` claim, or `None` if there is no `exp` claim

        Raises:
            AttributeError: if the `exp` claim cannot be parsed to a date
        """
        exp = self.get_claim("exp")
        if not exp:
            return None
        try:
            exp_dt = datetime.fromtimestamp(exp, tz=timezone.utc)
            return exp_dt
        except (TypeError, OSError):
            raise AttributeError("invalid `exp `claim", exp)

    @cached_property
    def issued_at(self) -> Optional[datetime]:
        """Get the "Issued At" (iat) date from this token.

        Returns:
          a `datetime` initialized from the `iat` claim, or `None` if there is no `iat` claim

        Raises:
            AttributeError: if the `iss` claim cannot be parsed to a date
        """
        iat = self.get_claim("iat")
        if not iat:
            return None
        try:
            iat_dt = datetime.fromtimestamp(iat, tz=timezone.utc)
            return iat_dt
        except (TypeError, OSError):
            raise AttributeError("invalid `iat `claim", iat)

    @cached_property
    def not_before(self) -> Optional[datetime]:
        """Get the "Not Before" (nbf) date from this token.

        Returns:
          a `datetime` initialized from the `nbf` claim, or `None` if there is no `nbf` claim

        Raises:
            AttributeError: if the `nbf` claim cannot be parsed to a date
        """
        nbf = self.get_claim("nbf")
        if not nbf:
            return None
        try:
            nbf_dt = datetime.fromtimestamp(nbf, tz=timezone.utc)
            return nbf_dt
        except (TypeError, OSError):
            raise AttributeError("invalid `nbf `claim", nbf)

    @cached_property
    def issuer(self) -> Optional[str]:
        """Get the Issuer (iss) claim from this token.

        Returns:
          the issuer, as `str`, or `None` if there is no `ìss` claim

        Raises:
            AttributeError: if the `ìss` claim value is not a string
        """
        iss = self.get_claim("iss")
        if iss is None or isinstance(iss, str):
            return iss
        raise AttributeError("iss has an unexpected type", type(iss))

    @cached_property
    def audiences(self) -> Optional[List[str]]:
        """Get the audience(s) (aud) claim from this token.

        If this token has a single audience, this will return a `list` anyway.

        Returns:
            the list of audiences from this token, from the `aud` claim.

        Raises:
            AttributeError: if the audience is an unexpected type
        """
        aud = self.get_claim("aud")
        if aud is None:
            return None
        if isinstance(aud, str):
            return [aud]
        if isinstance(aud, list):
            return aud
        raise AttributeError("aud has an unexpected type", type(aud))

    @cached_property
    def subject(self) -> Optional[str]:
        """Get the Subject (sub) from this token claims.

        Returns:
          the subject, as `str`, or `None` if there is no `sub` claim

        Raises:
            AttributeError: if the `sub` value is not a string
        """
        sub = self.get_claim("sub")
        if sub is None or isinstance(sub, str):
            return sub
        raise AttributeError("sub has an unexpected type", type(sub))

    @cached_property
    def jwt_token_id(self) -> Optional[str]:
        """Get the JWT Token ID (jti) from this token claims.

        Returns:
          the token identifier, as `str`, or `None` if there is no `jti` claim

        Raises:
          AttributeError: if the `jti` value is not a string
        """
        jti = self.get_claim("jti")
        if jti is None or isinstance(jti, str):
            return jti
        raise AttributeError("jti has an unexpected type", type(jti))

    @cached_property
    def alg(self) -> Optional[str]:
        """Get the signature algorithm from the header.

        Returns:
          the token signing alg, from the `alg` header

        Raises:
            AttributeError: if the alg is not a string
        """
        alg = self.get_header("alg")
        if alg is None or isinstance(alg, str):
            return alg
        raise AttributeError("alg has an unexpected type", type(alg))

    @cached_property
    def kid(self) -> Optional[str]:
        """Get the Key ID (kid) from the JWT header.

        Returns:
          the token signing key id, from the `kid` header

        Raises:
            AttributeError: if the kid is present but is not a string
        """
        kid = self.get_header("kid")
        if kid is None or isinstance(kid, str):
            return kid
        raise AttributeError("kid has an unexpected type", type(kid))

    def get_claim(self, key: str, default: Any = None) -> Any:
        """Get a claim from this Jwt.

        Args:
          key: the claim name.
          default: a default value if the claim is not found

        Returns:
          the claim value if found, or `default` if not found
        """
        return self.claims.get(key, default)

    def __getitem__(self, item: str) -> Any:
        """Allow claim access with subscription.

        Args:
          item: the claim name

        Returns:
         the claim value
        """
        value = self.get_claim(item)
        if value is None:
            raise KeyError(item)
        return value

    def __getattr__(self, item: str) -> Any:
        """Allow claim access as attributes.

        Args:
            item: the claim name

        Returns:
            the claim value
        """
        value = self.get_claim(item)
        if value is None:
            raise AttributeError(item)
        return value

    def __str__(self) -> str:
        """Return the Jwt serialized value, as `str`.

        Returns:
            the serialized token value.
        """
        return self.value.decode()

    def __bytes__(self) -> bytes:
        """Return the Jwt serialized value, as `bytes`.

        Returns:
            the serialized token value.
        """
        return self.value

    def validate(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
        *,
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
        issuer: Optional[str] = None,
        audience: Union[None, str] = None,
        check_exp: bool = True,
        **kwargs: Any,
    ) -> None:
        """Validate a `SignedJwt` signature and expected claims.

        This verifies the signature using the provided `jwk` and `alg`, then checks the token issuer, audience and expiration date.
        This can also check custom claims using extra `kwargs`, whose values can be:

        - a static value (`str`, `int`, etc.): the value from the token will be compared "as-is"
        - a callable, taking the claim value as parameter: if that callable returns `True`, the claim is considered as valid

        Args:
          jwk: the signing key to use to verify the signature.
          alg: the signature alg to use to verify the signature.
          algs: allowed signature algs, if several
          issuer: the expected issuer for this token.
          audience: the expected audience for this token.
          check_exp: ìf `True` (default), check that the token is not expired.
          **kwargs: additional claims to check

        Returns:
          Raises exceptions if any validation check fails.

        Raises:
          InvalidSignature: if the signature is not valid
          InvalidClaim: if a claim doesn't validate
          ExpiredJwt: if the expiration date is passed
        """
        if not self.verify_signature(jwk, alg, algs):
            raise InvalidSignature("Signature is not valid.")

        if issuer is not None:
            if self.issuer != issuer:
                raise InvalidClaim("iss", "Unexpected issuer", self.issuer)

        if audience is not None:
            if self.audiences is None or audience not in self.audiences:
                raise InvalidClaim("aud", "Unexpected audience", self.audiences)

        if check_exp:
            expired = self.is_expired()
            if expired is True:
                raise ExpiredJwt(f"This token expired at {self.expires_at}")
            elif expired is None:
                raise InvalidClaim("exp", "This token misses a 'exp' claim.")

        for key, value in kwargs.items():
            claim = self.get_claim(key)
            if callable(value):
                if not value(claim):
                    raise InvalidClaim(
                        key,
                        f"value of claim {key} doesn't validate with the provided validator",
                        claim,
                    )
            elif claim != value:
                raise InvalidClaim(key, f"unexpected value for claim {key}", claim)



"""This module implements RSA signature algorithms."""

from binapy import BinaPy
from cryptography import exceptions
from cryptography.hazmat.primitives import asymmetric, hashes
from cryptography.hazmat.primitives.asymmetric import padding

from ..base import BaseAsymmetricAlg, BaseSignatureAlg


class BaseRSASigAlg(
    BaseAsymmetricAlg[asymmetric.rsa.RSAPrivateKey, asymmetric.rsa.RSAPublicKey],
    BaseSignatureAlg,
):
    """Base class for RSA based signature algorithms."""

    hashing_alg: hashes.HashAlgorithm
    padding_alg: padding.AsymmetricPadding = padding.PKCS1v15()
    min_key_size: int = 2048

    private_key_class = asymmetric.rsa.RSAPrivateKey
    public_key_class = asymmetric.rsa.RSAPublicKey

    def sign(self, data: bytes) -> BinaPy:
        """Sign arbitrary data.

        Args:
          data: the data to sign

        Returns:
            the generated signature

        Raises:
            NotImplementedError: for algorithms that are considered insecure, only signature verification is available
            PrivateKeyRequired: if the configured key is not private
        """
        if self.read_only:
            raise NotImplementedError
        with self.private_key_required() as key:
            return BinaPy(key.sign(data, self.padding_alg, self.hashing_alg))

    def verify(self, data: bytes, signature: bytes) -> bool:
        """Verify a signature against some data.

        Args:
          data: the data to verify
          signature: the signature

        Returns:
            `True` if the signature is valid, `False` otherwise
        """
        with self.public_key_required() as key:
            try:
                key.verify(
                    signature,
                    data,
                    self.padding_alg,
                    self.hashing_alg,
                )
                return True
            except exceptions.InvalidSignature:
                return False


class RS256(BaseRSASigAlg):  # noqa: D415
    """RSASSA-PKCS1-v1_5 using SHA-256"""

    name = "RS256"
    description = __doc__
    hashing_alg = hashes.SHA256()


class RS384(BaseRSASigAlg):  # noqa: D415
    """RSASSA-PKCS1-v1_5 using SHA-384"""

    name = "RS384"
    description = __doc__
    hashing_alg = hashes.SHA384()


class RS512(BaseRSASigAlg):  # noqa: D415
    """RSASSA-PKCS1-v1_5 using SHA-256"""

    name = "RS512"
    description = __doc__
    hashing_alg = hashes.SHA512()


class PS256(BaseRSASigAlg):  # noqa: D415
    """RSASSA-PSS using SHA-256 and MGF1 with SHA-256"""

    name = "PS256"
    description = __doc__
    hashing_alg = hashes.SHA256()
    padding_alg = padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=256 // 8)


class PS384(BaseRSASigAlg):  # noqa: D415
    """RSASSA-PSS using SHA-384 and MGF1 with SHA-384"""

    name = "PS384"
    description = __doc__
    hashing_alg = hashes.SHA384()
    padding_alg = padding.PSS(mgf=padding.MGF1(hashes.SHA384()), salt_length=384 // 8)


class PS512(BaseRSASigAlg):  # noqa: D415
    """RSASSA-PSS using SHA-512 and MGF1 with SHA-512"""

    name = "PS512"
    description = __doc__
    hashing_alg = hashes.SHA512()
    padding_alg = padding.PSS(mgf=padding.MGF1(hashes.SHA512()), salt_length=512 // 8)


class RS1(BaseRSASigAlg):  # noqa: D415
    """RSASSA-PKCS1-v1_5 with SHA-1"""

    name = "RS1"
    description = __doc__
    hashing_alg = hashes.SHA1()
    read_only = True



"""Common fixtures for all tests."""

import pytest

from jwskate import Jwk


@pytest.fixture()
def issuer() -> str:
    """Return a valid issuer."""
    return "https://test.com"


@pytest.fixture()
def kid() -> str:
    """Return a valid 'kid' (Key ID)."""
    return "JWK-ABCD"


@pytest.fixture()
def private_jwk(kid: str) -> Jwk:
    """Return a private Jwk, usable for signing or decryption."""
    return Jwk(
        {
            "kty": "RSA",
            "kid": kid,
            "alg": "RS256",
            "n": "2jgK-5aws3_fjllgnAacPkwjbz3RCeAHni1pcHvReuTgk9qEiTmXWJiSS_F20VeI1zEwFM36e836ROCyOQ8cjjaPWpdzCajWC0koY7X8MPhZbdoSptOmDBseRCyYqmeMCp8mTTOD6Cs43SiIYSMNlPuio89qjf_4u32eVF_5YqOGtwfzC4p2NUPPCxpljYpAcf2BBG1tRX1mY4WP_8zwmx3ZH7Sy0V_fXI46tzDqfRXdMhHW7ARJAnEr_EJhlMgUaM7FUQKUNpi1ZdeeLxYv44eRx9-Roy5zTG1b0yRuaKaAG3559572quOcxISZzK5Iy7BhE7zxVa9jabEl-Y1Daw",
            "e": "AQAB",
            "d": "XCtpsCRQ1DBBm51yqdQ88C82lEjW30Xp0cy6iVEzBKZhmPGmI1PY8gnXWQ5PMlK3sLTM6yypDNvORoNlo6YXWJYA7LGlXEIczj2DOsJmF8T9-OEwGZixvNFDcmYnwWnlA6N_CQKmR0ziQr9ZAzZMCU5Tvr7f8cRZKdAALQEwk5FYpLnEbXOBduJtY9x2kddJSCJwRaEJhx0fG_pJAO3yLUZBY20dZK8UrxDoCgB9eiZV3N4uWGt367r1MDdaxGY6l6bC1HZCHkttBuTxfSUMCgooZevdU6ThQNpFrwZNY3KoP-OksEdqMs-neecfk_AQREkubDW2VPNFnaVEa38BKQ",
            "p": "8QNZGwUINpkuZi8l2ZfQzKVeOeNe3aQ7UW0wperM-63DFEJDRO1UyNC1n6yeo8_RxPZKSTlr6xZDoilQq23mopeF6O0ZmYz6E2VWJuma65V-A7tB-6xjqUXPlSkCNA6Ia8kMeCmNpKs0r0ijTBf_2y2GSsNH4EcP7XzcDEeJIh0",
            "q": "58nWgg-qRorRddwKM7qhLxJnEDsnCiYhbKJrP78OfBZ-839bNRvL5D5sfjJqxcKMQidgpYZVvVNL8oDEywcC5T7kKW0HK1JUdYiX9DuI40Mv9WzXQ8B8FBjp5wV4IX6_0KgyIiyoUiKpVHBvO0YFPUYuk0Ns4H9yEws93RWwhSc",
            "dp": "zFsLZcaphSnzVr9pd4urhqo9MBZjbMmBZnSQCE8ECe729ymMQlh-SFv3dHF4feuLsVcn-9iNceMJ6-jeNs1T_s89wxevWixYKrQFDa-MJW83T1CrDQvJ4VCJR69i5-Let43cXdLWACcO4AVWOQIsdpquQJw-SKPYlIUHS_4n_90",
            "dq": "fP79rNnhy3TlDBgDcG3-qjHUXo5nuTNi5wCXsaLInuZKw-k0OGmrBIUdYNizd744gRxXJCxTZGvdEwOaHJrFVvcZd7WSHiyh21g0CcNpSJVc8Y8mbyUIRJZC3RC3_egqbM2na4KFqvWCN0UC1wYloSuNxmCgAFj6HYb8b5NYxBU",
            "qi": "hxXfLYgwrfZBvZ27nrPsm6mLuoO-V2rKdOj3-YDJzf0gnVGBLl0DZbgydZ8WZmSLn2290mO_J8XY-Ss8PjLYbz3JXPDNLMJ-da3iEPKTvh6OfliM_dBxhaW8sq5afLMUR0H8NeabbWkfPz5h0W11CCBYxsyPC6CzniFYCYXfByU",
        }
    )



from jwskate import ECJwk


def test_jwk_ec_generate() -> None:
    jwk = ECJwk.generate(kid="myeckey")
    assert jwk.kty == "EC"
    assert jwk.kid == "myeckey"
    assert jwk.crv == "P-256"
    assert "x" in jwk
    assert "y" in jwk
    assert "d" in jwk

    public_jwk = jwk.public_jwk()
    assert public_jwk.kty == "EC"
    assert public_jwk.kid == "myeckey"
    assert public_jwk.crv == "P-256"
    assert "x" in public_jwk
    assert "y" in public_jwk

    assert jwk.supported_encryption_algorithms() == []



"""This module implement JWS signatures."""

from __future__ import annotations

from typing import Any, Dict, Iterable, Mapping, Optional, Type, TypeVar, Union

from backports.cached_property import cached_property
from binapy import BinaPy

from jwskate.jwk.base import BaseJsonDict, Jwk

S = TypeVar("S", bound="JwsSignature")


class JwsSignature(BaseJsonDict):
    """Represents a JWS Signature.

    A JWS Signature has

     - a protected header (as a JSON object)
     - a signature value (as raw data)
     - an unprotected header (as arbitrary JSON data)
     - optional extra JSON attributes
    """

    @classmethod
    def from_parts(
        cls: Type[S],
        protected: Mapping[str, Any],
        signature: bytes,
        header: Optional[Any],
        **kwargs: Any,
    ) -> S:
        """Initialize a JwsSignature based on the provided parts.

        Args:
          protected: the protected headers, as a key: value mapping
          signature: the raw signature value
          header: the unprotected header, if any
          **kwargs: extra attributes, if any

        Returns:
            A `JwsSignature` based on the provided parts.
        """
        content = dict(
            kwargs,
            protected=BinaPy.serialize_to("json", protected).to("b64u").ascii(),
            signature=BinaPy(signature).to("b64u").ascii(),
        )
        if header is not None:
            content["header"] = header
        return cls(content)

    @cached_property
    def protected(self) -> Dict[str, Any]:
        """The protected header.

        Returns:
            the protected headers, as a `dict`.

        Raises:
            AttributeError: if this signature doesn't have protected headers.
        """
        protected = self.get("protected")
        if protected is None:
            raise AttributeError("This Jws JSON does not contain a 'protected' member")
        return BinaPy(protected).decode_from("b64u").parse_from("json")  # type: ignore

    @property
    def header(self) -> Any:
        """The unprotected header, unaltered.

        Returns:
            The unprotected header
        """
        return self.get("header")

    @cached_property
    def signature(self) -> bytes:
        """The raw signature.

        Returns:
            The raw signed data, unencoded

        Raises:
            AttributeError: if no 'signature' member is present
        """
        signature = self.get("signature")
        if signature is None:
            raise AttributeError("This Jws JSON does not contain a 'signature' member")
        return BinaPy(signature).decode_from("b64u")

    @classmethod
    def sign(
        cls: Type[S],
        payload: bytes,
        jwk: Union[Jwk, Dict[str, Any]],
        alg: Optional[str] = None,
        extra_protected_headers: Optional[Mapping[str, Any]] = None,
        header: Optional[Any] = None,
        **kwargs: Any,
    ) -> S:
        """Sign a payload and return the generated JWS signature.

        Args:
          payload: the raw data to sign
          jwk: the signature key to use
          alg: the signature algorithm to use
          extra_protected_headers: additional protected headers to include, if any
          header: the unprotected header, if any.
          **kwargs: additional members to include in this signature

        Returns:
            The generated signature.
        """
        jwk = Jwk(jwk)

        headers = dict(extra_protected_headers or {}, alg=alg)
        kid = jwk.get("kid")
        if kid:
            headers["kid"] = kid

        signed_part = JwsSignature.assemble_signed_part(headers, payload)
        signature = jwk.sign(signed_part, alg=alg)
        return cls.from_parts(
            protected=headers, signature=signature, header=header, **kwargs
        )

    @classmethod
    def assemble_signed_part(
        cls, headers: Dict[str, Any], payload: Union[bytes, str]
    ) -> bytes:
        """Assemble the protected header and payload to sign, as specified in.

        [RFC7515
        $5.1](https://datatracker.ietf.org/doc/html/rfc7515#section-5.1).

        Args:
          headers: the protected headers
          payload: the raw payload to sign

        Returns:
            the raw data to sign
        """
        return b".".join(
            (
                BinaPy.serialize_to("json", headers).to("b64u"),
                BinaPy(payload).to("b64u"),
            )
        )

    def verify(
        self,
        payload: bytes,
        jwk: Union[Jwk, Dict[str, Any]],
        *,
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
    ) -> bool:
        """Verify this signature against the given payload using the provided key.

        Args:
          payload: the raw payload
          jwk: the validation key to use
          alg: the signature alg t if only 1 is allowed
          algs: the allowed signature algs, if there are several

        Returns:
            `True` if the signature is verifier, `False` otherwise
        """
        jwk = Jwk(jwk)
        signed_part = self.assemble_signed_part(self.protected, payload)
        return jwk.verify(signed_part, self.signature, alg=alg, algs=algs)



"""This module contains several utilities for algorithmic agility."""

import warnings
from typing import Any, Iterable, List, Mapping, Optional, Type, TypeVar

from jwskate.jwa import BaseAlg


class UnsupportedAlg(ValueError):
    """Raised when an UnsupportedAlg is requested."""


class ExpectedAlgRequired(ValueError):
    """Raised when the expected signature alg(s) must be provided."""


T = TypeVar("T", bound=Type[BaseAlg])


def select_alg(
    jwk_alg: Optional[str], alg: Optional[str], supported_algs: Mapping[str, T]
) -> T:
    """Given an alg parameter from a JWK, and/or a user-specified alg, return the alg to use.

    This checks the coherency between the user specified `alg` and the `jwk_alg`, and will emit a warning
    if the user specified alg is different from the `jwk_alg`.

    Args:
      jwk_alg: the alg from the JWK, if any
      alg: a user specified alg
      supported_algs: a mapping of supported alg names to alg wrapper

    Returns:
      the alg to use

    Raises:
        UnsupportedAlg: if the requested alg is not supported
    """
    choosen_alg: str
    if jwk_alg is not None:
        if alg is not None:
            if jwk_alg != alg:
                warnings.warn(
                    "This key has an 'alg' parameter, you should use that alg for each operation."
                )
            choosen_alg = alg
        else:
            choosen_alg = jwk_alg
    elif alg is not None:
        choosen_alg = alg
    else:
        raise ExpectedAlgRequired(
            "This key doesn't have an 'alg' parameter, so you need to provide the expected signing alg(s) for each operation."
        )

    try:
        return supported_algs[choosen_alg]
    except KeyError:
        raise UnsupportedAlg(
            f"Alg {choosen_alg} is not supported. Supported algs: {list(supported_algs)}."
        )


def select_algs(
    jwk_alg: Optional[str],
    alg: Optional[str],
    algs: Optional[Iterable[str]],
    supported_algs: Mapping[str, T],
) -> List[T]:
    """Given an alg parameter from a JWK, and/or a user-specified alg, and/or a user specified list of useable algs, return a list of algorithms.

    This method is typically used to get the list of possible algs when checking a signature.

    Args:
      jwk_alg: the alg from the JWK, if any
      alg: a user specified alg to use
      algs: a user specified list of algs to use, if several are allowed
      supported_algs: a mapping of alg names to alg wrappers

    Returns:
      a list of possible algs to check

    Raises:
        ValueError: if both 'alg' and 'algs' parameters are used
        UnsupportedAlg: if none of the requested alg are supported

    Warnings:
        if the requested 'alg' is different that the 'jwk_alg', or the 'jwk_alg' is not in the 'algs'
    """
    if alg and algs:
        raise ValueError("Please use either parameter 'alg' or 'algs', not both.")

    if jwk_alg is not None:
        if alg and alg != jwk_alg:
            warnings.warn(
                "This key has an 'alg' parameter, you should use that alg for each operation."
            )
        if algs and jwk_alg not in algs:
            warnings.warn(
                "This key has an 'alg' parameter, you should use that alg for each operation."
            )

    possible_algs: List[str] = []
    if alg:
        possible_algs = [alg]
    elif algs:
        possible_algs = list(algs)
    elif jwk_alg:
        possible_algs = [jwk_alg]

    if possible_algs:
        possible_supported_algs = [
            supported_algs[alg] for alg in possible_algs if alg in supported_algs
        ]
        if possible_supported_algs:
            return possible_supported_algs
        else:
            raise UnsupportedAlg(
                f"None of the user-specified alg(s) are supported. {possible_algs}"
            )

    raise ExpectedAlgRequired(
        "This key doesn't have an 'alg' parameter, so you need to provide the expected signing alg(s) for each operation."
    )


KEY_PARAMS_FOR_ALG: Mapping[str, Mapping[str, Any]] = {
    "HS256": {"kty": "oct", "key_size": 256, "use": "sig", "key_ops": ["sign"]},
    "HS384": {"kty": "oct", "key_size": 384, "use": "sig", "key_ops": ["sign"]},
    "HS512": {"kty": "oct", "key_size": 512, "use": "sig", "key_ops": ["sign"]},
    "RS256": {"kty": "RSA", "key_size": 2048, "use": "sig", "key_ops": ["sign"]},
    "RS384": {"kty": "RSA", "key_size": 3064, "use": "sig", "key_ops": ["sign"]},
    "RS512": {"kty": "RSA", "key_size": 4096, "use": "sig", "key_ops": ["sign"]},
    "PS256": {"kty": "RSA", "key_size": 2048, "use": "sig", "key_ops": ["sign"]},
    "PS384": {"kty": "RSA", "key_size": 3064, "use": "sig", "key_ops": ["sign"]},
    "PS512": {"kty": "RSA", "key_size": 4096, "use": "sig", "key_ops": ["sign"]},
    "ES256": {"kty": "EC", "crv": "P-256", "use": "sig", "key_ops": ["sign"]},
    "ES384": {"kty": "EC", "crv": "P-384", "use": "sig", "key_ops": ["sign"]},
    "ES512": {"kty": "EC", "crv": "P-521", "use": "sig", "key_ops": ["sign"]},
    "RSA1_5": {
        "kty": "RSA",
        "key_size": 2048,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "RSA-OAEP": {
        "kty": "RSA",
        "key_size": 3064,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "RSA-OAEP-256": {
        "kty": "RSA",
        "key_size": 4096,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "RSA-OAEP-384": {
        "kty": "RSA",
        "key_size": 4096,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "RSA-OAEP-512": {
        "kty": "RSA",
        "key_size": 4096,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "A128KW": {"kty": "oct", "key_size": 128, "use": "enc", "key_ops": ["wrapKey"]},
    "A192KW": {"kty": "oct", "key_size": 192, "use": "enc", "key_ops": ["wrapKey"]},
    "A256KW": {"kty": "oct", "key_size": 256, "use": "enc", "key_ops": ["wrapKey"]},
    "A128GCMKW": {
        "kty": "oct",
        "key_size": 128,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "A192GCMKW": {
        "kty": "oct",
        "key_size": 192,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "A256GCMKW": {
        "kty": "oct",
        "key_size": 256,
        "use": "enc",
        "key_ops": ["wrapKey"],
    },
    "ECDH-ES": {"kty": "EC", "crv": "X25519", "use": "enc", "key_ops": ["wrapKey"]},
    "A128CBC-HS256": {
        "kty": "oct",
        "key_size": 256,
        "use": "enc",
        "key_ops": ["encrypt"],
    },
    "A192CBC-HS384": {
        "kty": "oct",
        "key_size": 384,
        "use": "enc",
        "key_ops": ["encrypt"],
    },
    "A256CBC-HS512": {
        "kty": "oct",
        "key_size": 512,
        "use": "enc",
        "key_ops": ["encrypt"],
    },
    "A128GCM": {
        "kty": "oct",
        "key_size": 128,
        "use": "enc",
        "key_ops": ["encrypt"],
    },
    "A192GCM": {
        "kty": "oct",
        "key_size": 192,
        "use": "enc",
        "key_ops": ["encrypt"],
    },
    "A256GCM": {
        "kty": "oct",
        "key_size": 256,
        "use": "enc",
        "key_ops": ["encrypt"],
    },
}


def get_kty_for_alg(alg: str) -> Mapping[str, Any]:
    """Given an alg identifier, return key parameters for a key that is suitable for use with that alg.

    Args:
        alg: a signature or keymanagement or encryption alg identifier

    Returns:
        a dict of key parameters
    """
    return KEY_PARAMS_FOR_ALG[alg]



"""This module implements AES based Key Management algorithms."""

from binapy import BinaPy
from cryptography.hazmat.primitives import keywrap

from ..base import BaseKeyManagementAlg, BaseSymmetricAlg


class BaseAesKeyWrap(BaseKeyManagementAlg, BaseSymmetricAlg):
    """Base class for AES KW algorithms."""

    key_size: int
    """Required AES key size in bits."""

    @classmethod
    def check_key(cls, key: bytes) -> None:
        """Check that a key is valid for usage with this algorithm.

        To be valid, a key must be `bytes` and be of appropriate length (128, 192 or 256 bits).

        Args:
          key: a key to check

        Raises:
            ValueError: if the key is not appropriate
        """
        if not isinstance(key, bytes) or len(key) * 8 != cls.key_size:
            raise ValueError(f"Key must be {cls.key_size} bits")

    def wrap_key(self, plainkey: bytes) -> BinaPy:
        """Wrap a key.

        Args:
          plainkey: the key to wrap.

        Returns:
          BinaPy: the wrapped key.
        """
        return BinaPy(keywrap.aes_key_wrap(self.key, plainkey))

    def unwrap_key(self, cipherkey: bytes) -> BinaPy:
        """Unwrap a key.

        Args:
          cipherkey: the wrapped key.

        Returns:
          BinaPy: the unwrapped key.
        """
        return BinaPy(keywrap.aes_key_unwrap(self.key, cipherkey))


class A128KW(BaseAesKeyWrap):
    """AES Key Wrap with default initial value using 128-bit key."""

    name = "A128KW"
    description = __doc__
    key_size = 128


class A192KW(BaseAesKeyWrap):
    """AES Key Wrap with default initial value using 192-bit key."""

    name = "A192KW"
    description = __doc__
    key_size = 192


class A256KW(BaseAesKeyWrap):
    """AES Key Wrap with default initial value using 256-bit key."""

    name = "A256KW"
    description = __doc__
    key_size = 256



"""This module contains the `JwtSigner` class."""

import uuid
from datetime import datetime
from typing import Any, Dict, Iterable, Optional, Union

from jwskate import Jwk

from .base import Jwt
from .signed import SignedJwt


class JwtSigner:
    """A helper class to easily sign JWTs with standardised claims.

    The standardised claims include:

       - `ìat`: issued at date
       - `exp`: expiration date
       - `nbf`: not before date:
       - `iss`: issuer identifier
       - `sub`: subject identifier
       - `aud`: audience identifier
       - `jti`: JWT token ID

    The issuer, signing keys, signing alg and default lifetime are
    defined at initialization time, so you only have to define the
    subject, audience and custom claims when calling `JwtSigner.sign()`.
    This can be used as an alternative to `Jwt.sign()` when a single
    issuer issues multiple tokens.
    """

    def __init__(
        self,
        issuer: str,
        jwk: Jwk,
        alg: Optional[str] = None,
        default_lifetime: int = 60,
        default_leeway: Optional[int] = None,
    ):
        """Initialize a `JwtSigner`.

        Args:
            issuer: the issuer string to use as `ìss` claim for signed tokens.
            jwk: the private Jwk to use to sign tokens.
            alg: the signing alg to use to sign tokens.
            default_lifetime: the default lifetime, in seconds, to use for claim `exp`. This can be overridden
        when calling `.sign()`
            default_leeway: the default leeway, in seconds, to use for claim `nbf`. If None, no `nbf` claim is
        included. This can be overridden when calling `.sign()`
        """
        self.issuer = issuer
        self.jwk = jwk
        self.alg = jwk.alg or alg
        self.default_lifetime = default_lifetime
        self.default_leeway = default_leeway

    def sign(
        self,
        subject: Optional[str] = None,
        audience: Union[str, Iterable[str], None] = None,
        extra_claims: Optional[Dict[str, Any]] = None,
        extra_headers: Optional[Dict[str, Any]] = None,
        lifetime: Optional[int] = None,
        leeway: Optional[int] = None,
    ) -> SignedJwt:
        """Sign a Jwt.

        Claim 'issuer' will have the value defined at initialization time. Claim `iat`, `nbf` and `exp` will reflect
        the current time when the token is signed. `exp` includes `lifetime` seconds in the future, and `nbf`
        includes `leeway` seconds in the past.

        Args:
          subject: the subject to include in claim `sub`. (Default value = None)
          audience: the audience identifier(s) to include in claim `aud`.
          extra_claims: additional claims to include in the signed token. (Default value = None)
          extra_headers: additional headers to include in the header part. (Default value = None)
          lifetime: lifetime, in seconds, to use for the `exp` claim. If None, use the default_lifetime defined at initialization time.
          leeway: leeway, in seconds, to use for the `nbf` claim. If None, use the default_leeway defined at initialization time.

        Returns:
          the resulting signed token.
        """
        now = int(datetime.now().timestamp())
        lifetime = lifetime or self.default_lifetime
        exp = now + lifetime
        leeway = leeway or self.default_leeway
        nbf = (now - leeway) if leeway is not None else None
        jti = self.generate_jti()
        extra_claims = extra_claims or {}
        claims = {
            key: value
            for key, value in dict(
                extra_claims,
                iss=self.issuer,
                aud=audience,
                sub=subject,
                iat=now,
                exp=exp,
                nbf=nbf,
                jti=jti,
            ).items()
            if value is not None
        }
        return Jwt.sign(claims, jwk=self.jwk, alg=self.alg, extra_headers=extra_headers)

    def generate_jti(self) -> str:
        """Generate Jwt Token ID (jti) values.

        Default uses UUID4. Can be overridden in subclasses.

        Returns:
            A unique value suitable for use as JWT Token ID (jti) claim.
        """
        return str(uuid.uuid4())



"""This module implements JWK representing RSA keys."""

from __future__ import annotations

from typing import Any, Optional, Union

from backports.cached_property import cached_property
from binapy import BinaPy
from cryptography.hazmat.primitives.asymmetric import rsa

from jwskate.jwa import (
    PS256,
    PS384,
    PS512,
    RS256,
    RS384,
    RS512,
    RsaEsOaep,
    RsaEsOaepSha256,
    RsaEsOaepSha384,
    RsaEsOaepSha512,
    RsaEsPcks1v1_5,
)

from .base import Jwk, JwkParameter


class RSAJwk(Jwk):
    """Represent a RSA Jwk, with `kty=RSA`."""

    KTY = "RSA"
    CRYPTOGRAPHY_KEY_CLASSES = (rsa.RSAPrivateKey, rsa.RSAPublicKey)

    PARAMS = {
        "n": JwkParameter("Modulus", is_private=False, is_required=True, kind="b64u"),
        "e": JwkParameter("Exponent", is_private=False, is_required=True, kind="b64u"),
        "d": JwkParameter(
            "Private Exponent", is_private=True, is_required=True, kind="b64u"
        ),
        "p": JwkParameter(
            "First Prime Factor", is_private=True, is_required=False, kind="b64u"
        ),
        "q": JwkParameter(
            "Second Prime Factor", is_private=True, is_required=False, kind="b64u"
        ),
        "dp": JwkParameter(
            "First Factor CRT Exponent", is_private=True, is_required=False, kind="b64u"
        ),
        "dq": JwkParameter(
            "Second Factor CRT Exponent",
            is_private=True,
            is_required=False,
            kind="b64u",
        ),
        "qi": JwkParameter(
            "First CRT Coefficient", is_private=True, is_required=False, kind="b64u"
        ),
        "oth": JwkParameter(
            "Other Primes Info", is_private=True, is_required=False, kind="unsupported"
        ),
    }

    SIGNATURE_ALGORITHMS = {
        sigalg.name: sigalg for sigalg in [RS256, RS384, RS512, PS256, PS384, PS512]
    }

    KEY_MANAGEMENT_ALGORITHMS = {
        keyalg.name: keyalg
        for keyalg in [
            RsaEsPcks1v1_5,
            RsaEsOaep,
            RsaEsOaepSha256,
            RsaEsOaepSha384,
            RsaEsOaepSha512,
        ]
    }

    @property
    def is_private(self) -> bool:  # noqa: D102
        return "d" in self

    @classmethod
    def from_cryptography_key(cls, cryptography_key: Any, **kwargs: Any) -> RSAJwk:
        """Initialize a Jwk from a `cryptography` RSA key.

        Args:
          cryptography_key: a `cryptography` RSA key
          **kwargs: additional members to include in the Jwk

        Returns:
            a RSAJwk initialized with the given key

        Raises:
            TypeError: if the given key type is not supported
        """
        if isinstance(cryptography_key, rsa.RSAPrivateKey):
            priv = cryptography_key.private_numbers()  # type: ignore[attr-defined]
            pub = cryptography_key.public_key().public_numbers()
            return cls.private(
                n=pub.n,
                e=pub.e,
                d=priv.d,
                p=priv.p,
                q=priv.q,
                dp=priv.dmp1,
                dq=priv.dmq1,
                qi=priv.iqmp,
            )
        elif isinstance(cryptography_key, rsa.RSAPublicKey):
            pub = cryptography_key.public_numbers()
            return cls.public(
                n=pub.n,
                e=pub.e,
            )
        else:
            raise TypeError("A RSAPrivateKey or a RSAPublicKey is required.")

    def _to_cryptography_key(self) -> Union[rsa.RSAPrivateKey, rsa.RSAPublicKey]:
        """Initialize a `cryptography` key based on this Jwk.

        Returns:
            a cryptography RSAPrivateKey or RSAPublicKey
        """
        if self.is_private:
            return rsa.RSAPrivateNumbers(
                self.first_prime_factor,
                self.second_prime_factor,
                self.private_exponent,
                self.first_factor_crt_exponent,
                self.second_factor_crt_exponent,
                self.first_crt_coefficient,
                rsa.RSAPublicNumbers(self.exponent, self.modulus),
            ).private_key()
        else:
            return rsa.RSAPublicNumbers(e=self.exponent, n=self.modulus).public_key()

    @classmethod
    def public(cls, n: int, e: int, **params: Any) -> RSAJwk:
        """Initialize a public RsaJwk from a modulus and an exponent.

        Args:
          n: the modulus
          e: the exponent
          **params: additional members to include in the Jwk

        Returns:
          a RsaJwk initialized from the provided parameters
        """
        return cls(
            dict(
                kty="RSA",
                n=BinaPy.from_int(n).to("b64u").ascii(),
                e=BinaPy.from_int(e).to("b64u").ascii(),
                **params,
            )
        )

    @classmethod
    def private(
        cls,
        n: int,
        e: int,
        d: int,
        p: Optional[int] = None,
        q: Optional[int] = None,
        dp: Optional[int] = None,
        dq: Optional[int] = None,
        qi: Optional[int] = None,
        **params: Any,
    ) -> RSAJwk:
        """Initializes a Private RsaJwk from its required parameters.

        Args:
          n: the modulus
          e: the exponent
          d: the private exponent
          p: the first prime factor
          q: the second prime factor
          dp: the first factor CRT exponent
          dq: the second factor CRT exponent
          qi: the first CRT coefficient
          **params: additional members to include in the Jwk

        Returns:
            a RSAJwk initialized from the given parameters
        """
        return cls(
            dict(
                kty="RSA",
                n=BinaPy.from_int(n).to("b64u").ascii(),
                e=BinaPy.from_int(e).to("b64u").ascii(),
                d=BinaPy.from_int(d).to("b64u").ascii(),
                p=BinaPy.from_int(p).to("b64u").ascii() if p is not None else None,
                q=BinaPy.from_int(q).to("b64u").ascii() if q is not None else None,
                dp=BinaPy.from_int(dp).to("b64u").ascii() if dp is not None else None,
                dq=BinaPy.from_int(dq).to("b64u").ascii() if dq is not None else None,
                qi=BinaPy.from_int(qi).to("b64u").ascii() if qi is not None else None,
                **params,
            )
        )

    @classmethod
    def generate(cls, key_size: int = 4096, **params: Any) -> RSAJwk:
        """Generates a new random private RSAJwk.

        Args:
          key_size: the key size to use for the generated key, in bits
          **params: additional members to include in the Jwk

        Returns:
          a generated RSAJwk
        """
        private_key = rsa.generate_private_key(65537, key_size=key_size)
        pn = private_key.private_numbers()
        return cls.private(
            n=pn.public_numbers.n,
            e=pn.public_numbers.e,
            d=pn.d,
            p=pn.p,
            q=pn.q,
            dp=pn.dmp1,
            dq=pn.dmq1,
            qi=pn.iqmp,
            **params,
        )

    @cached_property
    def modulus(self) -> int:
        """Returns the modulus from this Jwk.

        Returns:
            the key modulus (from parameter `n`)
        """
        return BinaPy(self.n).decode_from("b64u").to_int()

    @cached_property
    def exponent(self) -> int:
        """Returns the exponent from this Jwk.

        Returns:
            the key exponent (from parameter `e`)
        """
        return BinaPy(self.e).decode_from("b64u").to_int()

    @cached_property
    def private_exponent(self) -> int:
        """Returns the private exponent from this Jwk.

        Returns:
            the key private exponent (from parameter `d`)
        """
        return BinaPy(self.d).decode_from("b64u").to_int()

    @cached_property
    def first_prime_factor(self) -> int:
        """Returns the first prime factor from this Jwk.

        Returns:
            the first prime factor (from parameter `p`)
        """
        return BinaPy(self.p).decode_from("b64u").to_int()

    @cached_property
    def second_prime_factor(self) -> int:
        """Returns the second prime factor from this Jwk.

        Returns:
            the second prime factor (from parameter `q`)
        """
        return BinaPy(self.q).decode_from("b64u").to_int()

    @cached_property
    def first_factor_crt_exponent(self) -> int:
        """Returns the first factor CRT exponent from this Jwk.

        Returns:
            the first factor CRT coefficient (from parameter `dp`)
        """
        return BinaPy(self.dp).decode_from("b64u").to_int()

    @cached_property
    def second_factor_crt_exponent(self) -> int:
        """Returns the second factor CRT exponent from this Jwk.

        Returns:
            the second factor CRT coefficient (from parameter `dq`)
        """
        return BinaPy(self.dq).decode_from("b64u").to_int()

    @cached_property
    def first_crt_coefficient(self) -> int:
        """Returns the first CRT coefficient from this Jwk.

        Returns:
            the first CRT coefficient (from parameter `qi`)
        """
        return BinaPy(self.qi).decode_from("b64u").to_int()



"""This module exposes all the Signature algorithms available from `jwskate`."""

from .ec import ES256, ES256K, ES384, ES512
from .eddsa import EdDsa
from .hmac import HS256, HS384, HS512
from .rsa import PS256, PS384, PS512, RS256, RS384, RS512

__all__ = [
    "HS256",
    "HS384",
    "HS512",
    "ES256",
    "ES256K",
    "ES384",
    "ES512",
    "EdDsa",
    "PS256",
    "PS384",
    "PS512",
    "RS256",
    "RS384",
    "RS512",
]
